from datetime import datetime
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import pendulum

from tecton_core import conf
from tecton_core import data_types
from tecton_core import errors
from tecton_core import feature_definition_wrapper
from tecton_core import specs
from tecton_core import time_utils
from tecton_core.compute_mode import ComputeMode
from tecton_core.embeddings.config import TextEmbeddingInferenceConfig
from tecton_core.embeddings.config import TextEmbeddingModel
from tecton_core.feature_set_config import FeatureDefinitionAndJoinConfig
from tecton_core.feature_set_config import FeatureSetConfig
from tecton_core.feature_set_config import find_dependent_feature_set_items
from tecton_core.pipeline_common import get_time_window_from_data_source_node
from tecton_core.query import compaction_utils
from tecton_core.query.dialect import Dialect
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import AddAnchorTimeColumnsForSawtoothIntervalsNode
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import AddBooleanPartitionColumnsNode
from tecton_core.query.nodes import AddDurationNode
from tecton_core.query.nodes import AddEffectiveTimestampNode
from tecton_core.query.nodes import AddRetrievalAnchorTimeNode
from tecton_core.query.nodes import AddUniqueIdNode
from tecton_core.query.nodes import AdjustAnchorTimeToWindowEndNode
from tecton_core.query.nodes import AggregationSecondaryKeyExplodeNode
from tecton_core.query.nodes import AggregationSecondaryKeyRollupNode
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.query.nodes import AsofJoinInputContainer
from tecton_core.query.nodes import AsofJoinNode
from tecton_core.query.nodes import AsofJoinSawtoothAggNode
from tecton_core.query.nodes import AsofSecondaryKeyExplodeNode
from tecton_core.query.nodes import ConvertEpochToTimestampNode
from tecton_core.query.nodes import ConvertTimestampToUTCNode
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import DeriveValidityPeriodNode
from tecton_core.query.nodes import ExplodeEventsAndSelectDistinctNode
from tecton_core.query.nodes import ExplodeTimestampByTimeWindowsNode
from tecton_core.query.nodes import FeatureTimeFilterNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import InnerJoinOnRangeNode
from tecton_core.query.nodes import JoinNode
from tecton_core.query.nodes import MetricsCollectorNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import OnlineListAggNode
from tecton_core.query.nodes import OnlinePartialAggNodeV2
from tecton_core.query.nodes import PartialAggNode
from tecton_core.query.nodes import PythonDataNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import RespectFeatureStartTimeNode
from tecton_core.query.nodes import RespectTTLNode
from tecton_core.query.nodes import SelectDistinctNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import StreamWatermarkNode
from tecton_core.query.nodes import TakeLastRowNode
from tecton_core.query.nodes import TemporalBatchTableFormatNode
from tecton_core.query.nodes import TextEmbeddingInferenceNode
from tecton_core.query.nodes import TrimValidityPeriodNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.nodes import WildcardJoinNode
from tecton_core.query_consts import aggregation_group_id
from tecton_core.query_consts import aggregation_tile_id
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import default_case
from tecton_core.query_consts import effective_timestamp
from tecton_core.query_consts import exclusive_end_time
from tecton_core.query_consts import expiration_timestamp
from tecton_core.query_consts import inclusive_start_time
from tecton_core.query_consts import interval_end_time
from tecton_core.query_consts import interval_start_time
from tecton_core.query_consts import odfv_internal_staging_table
from tecton_core.query_consts import tecton_spine_row_id_col
from tecton_core.query_consts import tecton_unique_id_col
from tecton_core.query_consts import timestamp_plus_ttl
from tecton_core.query_consts import udf_internal
from tecton_core.query_consts import window_end_column_name
from tecton_core.schema import Schema
from tecton_core.specs import DataSourceSpec
from tecton_proto.args.pipeline__client_pb2 import DataSourceNode as ProtoDataSourceNode
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2


def _is_wildcard_join(fdw: feature_definition_wrapper.FeatureDefinitionWrapper, spine_node: NodeRef) -> bool:
    """Check whether this retrieval query needs to handle this FV with the "wildcard" join pattern.

    A feature view uses a wildcard join when:
      1. the FV has a wildcard join key
      2. that join key is not in the spine.

    NOTE: If the join key is in the spine, then we treat it like a standard FV (i.e. we include that spine column in the join).

    Wildcards are deprecated and only used by one customer. However, we need this code around until this functionality
    is removed from the product.
    """
    return fdw.wildcard_join_key is not None and fdw.wildcard_join_key not in spine_node.columns


def build_datasource_scan_node(
    dialect: Dialect,
    compute_mode: ComputeMode,
    ds: specs.DataSourceSpec,
    for_stream: bool,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> NodeRef:
    tree = DataSourceScanNode(
        dialect=dialect,
        compute_mode=compute_mode,
        ds=ds,
        ds_node=None,
        is_stream=for_stream,
        start_time=start_time,
        end_time=end_time,
    ).as_ref()
    return StagingNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=tree,
        staging_table_name=f"{ds.name}",
        query_tree_step=QueryTreeStep.DATA_SOURCE,
    ).as_ref()


def _get_ds_time_limits(
    feature_data_time_limits: Optional[pendulum.Period],
    schedule_interval: Optional[pendulum.Duration],
    data_source_node: ProtoDataSourceNode,
) -> Tuple[Optional[datetime], Optional[datetime]]:
    ds_time_limits = get_time_window_from_data_source_node(
        feature_data_time_limits, schedule_interval, data_source_node
    )
    if ds_time_limits:
        return ds_time_limits.start, ds_time_limits.end
    return None, None


def _build_datasource_input_querynodes(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    for_stream: bool,
    feature_data_time_limits: Optional[pendulum.Period] = None,
) -> Dict[str, NodeRef]:
    """
    Starting in FWV5, data sources of FVs with incremental backfills may contain transformations that are only
    correct if the data has been filtered to a specific range.
    """
    schedule_interval = fdw.get_tile_interval if fdw.is_temporal else None
    ds_inputs = feature_definition_wrapper.pipeline_to_ds_inputs(fdw.pipeline)

    input_querynodes = {}
    for input_name, node in ds_inputs.items():
        start_time, end_time = _get_ds_time_limits(feature_data_time_limits, schedule_interval, node)
        ds = fdw.fco_container.get_by_id_proto(node.virtual_data_source_id)
        assert isinstance(ds, DataSourceSpec)
        tree = DataSourceScanNode(
            dialect=dialect,
            compute_mode=compute_mode,
            ds=ds,
            ds_node=node,
            is_stream=for_stream,
            start_time=start_time,
            end_time=end_time,
        ).as_ref()
        input_querynodes[input_name] = StagingNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=tree,
            staging_table_name=f"{ds.name}",
            query_tree_step=QueryTreeStep.DATA_SOURCE,
        ).as_ref()
    return input_querynodes


def _get_stream_watermark(fdw: feature_definition_wrapper.FeatureDefinitionWrapper) -> Optional[str]:
    ds_inputs = feature_definition_wrapper.pipeline_to_ds_inputs(fdw.pipeline)
    for input_name, node in ds_inputs.items():
        ds_spec = fdw.fco_container.get_by_id_proto(node.virtual_data_source_id)
        assert isinstance(ds_spec, DataSourceSpec)
        if ds_spec.stream_source is not None:
            watermark_delay_threshold_seconds = ds_spec.stream_source.watermark_delay_threshold.total_seconds()
            # NOTE: we do not want to set an explicit '0 seconds' watermark as
            # that can lead to data loss (data source functions supports
            # user-specified watermark configuration in function).
            if watermark_delay_threshold_seconds:
                return f"{watermark_delay_threshold_seconds} seconds"
    return None


# build QueryTree that executes all transformations
def build_pipeline_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    for_stream: bool,
    feature_data_time_limits: Optional[pendulum.Period] = None,
) -> NodeRef:
    inputs_map = _build_datasource_input_querynodes(dialect, compute_mode, fdw, for_stream, feature_data_time_limits)
    base = FeatureViewPipelineNode(
        dialect=dialect,
        compute_mode=compute_mode,
        inputs_map=inputs_map,
        feature_definition_wrapper=fdw,
        feature_time_limits=feature_data_time_limits,
        check_view_schema=fdw.has_explicit_view_schema,
    ).as_ref()
    base = ConvertTimestampToUTCNode.for_feature_definition(
        dialect=dialect,
        compute_mode=compute_mode,
        fd=fdw,
        input_node=base,
    )

    if feature_data_time_limits:
        return FeatureTimeFilterNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=base,
            feature_data_time_limits=feature_data_time_limits,
            policy=fdw.time_range_policy,
            start_timestamp_field=fdw.timestamp_key,
            end_timestamp_field=fdw.timestamp_key,
        ).as_ref()
    return base


def build_snowflake_materialization_querytree(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    feature_data_time_limits: Optional[pendulum.Period] = None,
) -> NodeRef:
    materialization_node = build_materialization_querytree(
        dialect=Dialect.SNOWFLAKE,
        compute_mode=ComputeMode.SNOWFLAKE,
        fdw=fdw,
        for_stream=False,
        feature_data_time_limits=feature_data_time_limits,
        use_timestamp_key=True,
    )
    return materialization_node


def build_materialization_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    for_stream: bool,
    feature_data_time_limits: Optional[pendulum.Period] = None,
    include_window_end_time: bool = False,
    aggregation_anchor_time: Optional[datetime] = None,
    enable_feature_metrics: bool = False,
    use_timestamp_key: bool = False,
) -> NodeRef:
    """Builds a querytree to construct a dataframe for materialization.

    For example, WAFVs are partially aggregated, and BFVs are augmented with an anchor time column. The resulting
    dataframe can also be easily modified to be used for `fv.run`.

    Args:
        dialect: The SQL dialect
        compute_mode: Current compute mode
        fdw: The feature view to be materialized.
        for_stream: If True, the underlying data source is a streaming source.
        feature_data_time_limits: If set, the resulting features will be filtered with respect to these time limits.
        include_window_end_time: If True, a tile end time column with name "tile_end_time" will be included for WAFVs.
            Should only be set for WAFVs.
        aggregation_anchor_time: If set, it will be used as the offset for aggregations. Should only be set for WAFVs.
        enable_feature_metrics: If True, metrics will be collected on the querytree.
        use_timestamp_key: If True, the timestamp key will be used instead of _ANCHOR_TIME. This is used for Snowflake Compute only.
    """
    assert not for_stream or feature_data_time_limits is None, "Cannot run with time limits on a stream source"
    tree = build_pipeline_querytree(dialect, compute_mode, fdw, for_stream, feature_data_time_limits)
    if for_stream:
        watermark = _get_stream_watermark(fdw)
        if watermark:
            tree = StreamWatermarkNode(dialect, compute_mode, tree, fdw.time_key, watermark).as_ref()
    if enable_feature_metrics:
        tree = MetricsCollectorNode(dialect, compute_mode, tree).as_ref()

    if fdw.compaction_enabled and not for_stream:
        return _build_compaction_materialization_querytree(
            dialect=dialect,
            compute_mode=compute_mode,
            fdw=fdw,
            pipeline_tree=tree,
        )

    anchor_time_field = anchor_time()
    if fdw.is_temporal:
        tree = StagingNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=tree,
            staging_table_name=f"{fdw.name}",
            query_tree_step=QueryTreeStep.PIPELINE,
        ).as_ref()

        if len(fdw.fv_spec.embedding_features):
            tree = _add_text_embedding_inference_node(fdw=fdw, compute_mode=compute_mode, pipeline_tree=tree)

        # BFVs require an anchor time column, but SFVs do not.
        if not for_stream:
            assert not include_window_end_time, "Not supported window end time for temporal"
            if not use_timestamp_key:
                tree = AddAnchorTimeNode.for_feature_definition(dialect, compute_mode, fdw, tree)
    elif fdw.is_temporal_aggregate:
        end_column_name = window_end_column_name() if include_window_end_time else None
        tree = StagingNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=tree,
            staging_table_name=f"{fdw.name}",
            query_tree_step=QueryTreeStep.PIPELINE,
        ).as_ref()
        tree = PartialAggNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=tree,
            fdw=fdw,
            aggregation_tile_interval=fdw.get_tile_interval,
            window_start_column_name=anchor_time_field,
            window_end_column_name=end_column_name,
            aggregation_anchor_time=aggregation_anchor_time,
        ).as_ref()
        if use_timestamp_key:
            tree = ConvertEpochToTimestampNode(
                dialect, compute_mode, tree, {anchor_time(): fdw.get_feature_store_format_version}
            ).as_ref()
            tree = RenameColsNode(
                dialect, compute_mode, input_node=tree, mapping={anchor_time(): fdw.time_key}
            ).as_ref()
    else:
        msg = "unexpected FV type"
        raise Exception(msg)
    return tree


def _add_text_embedding_inference_node(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper, compute_mode: ComputeMode, pipeline_tree: NodeRef
) -> NodeRef:
    if compute_mode != ComputeMode.RIFT:
        msg = "text embeddings are only supported on Rift"
        raise ValueError(msg)

    qt_node = TextEmbeddingInferenceNode(
        dialect=Dialect.TORCH,
        compute_mode=ComputeMode.RIFT,
        input_node=pipeline_tree,
        inference_configs=tuple(
            TextEmbeddingInferenceConfig(
                input_column=e.input_column_name, output_column=e.output_column_name, model=TextEmbeddingModel(e.model)
            )
            for e in fdw.fv_spec.embedding_features
        ),
    )

    staging_node = StagingNode(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        input_node=qt_node.as_ref(),
        staging_table_name="model_inference",
        query_tree_step=QueryTreeStep.MODEL_INFERENCE,
    )

    # NOTE: we want to drop all columns that are not in the materialization_schema (i.e. indicates that we don't have the attribute set)
    m13n_col_names = set(fdw.materialization_schema.column_names())
    cols_to_drop = [col_name for col_name in fdw.view_schema.column_names() if col_name not in m13n_col_names]

    tree = RenameColsNode(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        input_node=staging_node.as_ref(),
        drop=cols_to_drop,
    )

    return tree.as_ref()


def _build_compaction_materialization_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    pipeline_tree: NodeRef,
) -> NodeRef:
    # Offline materialization query for fvs with compaction_enabled=True
    tree = StagingNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=pipeline_tree,
        staging_table_name=f"{fdw.name}",
        query_tree_step=QueryTreeStep.PIPELINE,
    ).as_ref()
    tree = ConvertTimestampToUTCNode.for_feature_definition(
        dialect=dialect, compute_mode=compute_mode, fd=fdw, input_node=tree
    )
    return AddAnchorTimeNode.for_feature_definition(dialect, compute_mode, fdw, tree)


def build_get_partial_aggregates_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    limits: pendulum.Period,
    entities: Optional[DataframeWrapper],
) -> NodeRef:
    qt = build_get_features(
        dialect=dialect,
        compute_mode=compute_mode,
        fdw=fdw,
        from_source=from_source,
        feature_data_time_limits=limits,
        include_window_end_time=True,
    )

    rename_mapping = {
        anchor_time(): interval_start_time(),
        window_end_column_name(): interval_end_time(),
    }

    qt = RenameColsNode(dialect, compute_mode, qt, mapping=rename_mapping).as_ref()

    qt = ConvertEpochToTimestampNode(
        dialect,
        compute_mode,
        qt,
        {
            interval_start_time(): fdw.get_feature_store_format_version,
            interval_end_time(): fdw.get_feature_store_format_version,
        },
    ).as_ref()

    if entities is not None:
        qt = _filter_entity_dataframe(dialect, compute_mode, qt, entities)

    return qt


def _build_get_features(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    feature_data_time_limits: Optional[pendulum.Period] = None,
    aggregation_anchor_time: Optional[datetime] = None,
    include_window_end_time: Optional[bool] = False,
) -> NodeRef:
    # NOTE: this is ideally the *only* place where we validate
    # from_source arguments. However, until Snowflake and Athena are migrated
    # to QueryTree, we also need validations to live in the interactive/unified
    # SDK.
    #
    # Behavior:
    #   from_source is True: force compute from source
    #   from_source is False: force compute from materialized data
    #   from_source is None: compute from materialized data if feature
    #       definition offline=True, otherwise compute from source
    if from_source is None:
        from_source = not fdw.materialization_enabled or not fdw.writes_to_offline_store

    if from_source is False:
        assert not aggregation_anchor_time, "aggregation anchor time is not allowed when fetching features from source"
        if not fdw.materialization_enabled or not fdw.writes_to_offline_store:
            raise errors.FV_NEEDS_TO_BE_MATERIALIZED(fdw.name)
        tree = OfflineStoreScanNode(
            dialect=dialect,
            compute_mode=compute_mode,
            feature_definition_wrapper=fdw,
            partition_time_filter=feature_data_time_limits,
        ).as_ref()
    else:
        if dialect == "athena":
            msg = "When Athena compute is enabled, features can only be read from the offline store. Please set from_source = False"
            raise errors.TectonAthenaValidationError(msg)
        # TODO(TEC-13005)
        # TODO(pooja): raise an appropriate error here for push source
        if fdw.is_incremental_backfill:
            raise errors.FV_BFC_SINGLE_FROM_SOURCE

        tree = build_materialization_querytree(
            dialect,
            compute_mode,
            fdw,
            for_stream=False,
            feature_data_time_limits=feature_data_time_limits,
            aggregation_anchor_time=aggregation_anchor_time,
            include_window_end_time=include_window_end_time,
        )
    return tree


def build_get_features(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    feature_data_time_limits: Optional[pendulum.Period] = None,
    aggregation_anchor_time: Optional[datetime] = None,
    include_window_end_time: Optional[bool] = False,
) -> NodeRef:
    tree = _build_get_features(
        dialect=dialect,
        compute_mode=compute_mode,
        fdw=fdw,
        from_source=from_source,
        feature_data_time_limits=feature_data_time_limits,
        aggregation_anchor_time=aggregation_anchor_time,
        include_window_end_time=include_window_end_time,
    )

    if fdw.compaction_enabled and fdw.is_temporal_aggregate:
        end_column_name = window_end_column_name() if include_window_end_time else None
        tree = PartialAggNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=tree,
            fdw=fdw,
            window_start_column_name=anchor_time(),
            aggregation_tile_interval=fdw.get_tile_interval_duration_for_offline_store,
            window_end_column_name=end_column_name,
            aggregation_anchor_time=aggregation_anchor_time,
        ).as_ref()

    return tree


def build_temporal_time_range_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    query_time_range: pendulum.Period,
    entities: Optional[DataframeWrapper],
) -> NodeRef:
    qt = build_get_features(
        dialect=dialect,
        compute_mode=compute_mode,
        fdw=fd,
        from_source=from_source,
        feature_data_time_limits=query_time_range,
    )
    qt = RenameColsNode(dialect, compute_mode, qt, drop=[anchor_time()]).as_ref()
    batch_schedule_seconds = 0 if fd.is_feature_table else fd.batch_materialization_schedule.in_seconds()

    qt = AddEffectiveTimestampNode(
        dialect,
        compute_mode,
        qt,
        timestamp_field=fd.timestamp_key,
        effective_timestamp_name=effective_timestamp(),
        batch_schedule_seconds=batch_schedule_seconds,
        data_delay_seconds=fd.online_store_data_delay_seconds,
        is_stream=fd.is_stream,
        is_temporal_aggregate=False,
    ).as_ref()

    if entities is not None:
        qt = _filter_entity_dataframe(dialect, compute_mode, qt, entities)

    qt = _filter_by_time_range(dialect, compute_mode, qt, query_time_range, fd.timestamp_key, fd.timestamp_key)
    return qt


def build_temporal_time_range_validity_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fd: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    query_time_range: pendulum.Period,
    lookback_time_range: pendulum.Period,
    entities: Optional[DataframeWrapper],
) -> NodeRef:
    qt = build_get_features(
        dialect=dialect,
        compute_mode=compute_mode,
        fdw=fd,
        from_source=from_source,
        feature_data_time_limits=lookback_time_range,
    )
    qt = RenameColsNode(dialect, compute_mode, qt, drop=[anchor_time()]).as_ref()
    batch_schedule_seconds = 0 if fd.is_feature_table else fd.batch_materialization_schedule.in_seconds()

    qt = AddEffectiveTimestampNode(
        dialect,
        compute_mode,
        qt,
        timestamp_field=fd.timestamp_key,
        effective_timestamp_name=effective_timestamp(),
        batch_schedule_seconds=batch_schedule_seconds,
        data_delay_seconds=fd.online_store_data_delay_seconds,
        is_stream=fd.is_stream,
        is_temporal_aggregate=False,
    ).as_ref()

    qt = TakeLastRowNode(
        dialect,
        compute_mode,
        input_node=qt,
        partition_by_columns=(*fd.join_keys, effective_timestamp()),
        order_by_column=fd.timestamp_key,
    ).as_ref()

    qt = DeriveValidityPeriodNode(dialect, compute_mode, qt, fd, effective_timestamp()).as_ref()

    if entities is not None:
        qt = _filter_entity_dataframe(dialect, compute_mode, qt, entities)

    qt = TrimValidityPeriodNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=qt,
        start=query_time_range.start,
        end=query_time_range.end,
    ).as_ref()

    return qt


def build_aggregated_time_range_validity_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    # query_time_range is the validated time range passed in the query
    query_time_range: pendulum.Period,
    # feature_data_time_limits is the aligned time range after taking the spine and aggregations into account
    feature_data_time_limits: pendulum.Period,
    entities: Optional[DataframeWrapper],
) -> NodeRef:
    partial_aggs = build_get_features(
        dialect,
        compute_mode,
        fdw,
        from_source,
        feature_data_time_limits=feature_data_time_limits,
    )

    if entities is not None:
        partial_aggs = _filter_entity_dataframe(dialect, compute_mode, partial_aggs, entities)

    spine = _build_internal_spine(
        dialect, compute_mode, fdw, partial_aggs, query_time_range=query_time_range, explode_anchor_time=True
    )

    # TODO(samantha): support sawtooth aggs here.
    full_aggregation_node = AsofJoinFullAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        spine=spine,
        partial_agg_node=partial_aggs,
        fdw=fdw,
        # Do not push down the timestamp if the spine is completely from partial agg, such as `ghf` and `run` with time
        # range.
        enable_spine_time_pushdown_rewrite=False,
        enable_spine_entity_pushdown_rewrite=False,
    ).as_ref()

    if fdw.aggregation_secondary_key:
        full_aggregation_node = AggregationSecondaryKeyRollupNode(
            dialect=dialect,
            compute_mode=compute_mode,
            full_aggregation_node=full_aggregation_node,
            fdw=fdw,
            group_by_columns=[*list(fdw.join_keys), anchor_time()],
        ).as_ref()

    qt = _filter_and_update_timestamp_columns(
        full_aggregation_node,
        dialect,
        compute_mode,
        fdw,
        show_effective_time=True,
    )

    qt = TakeLastRowNode(
        dialect,
        compute_mode,
        input_node=qt,
        partition_by_columns=(*fdw.join_keys, effective_timestamp()),
        order_by_column=fdw.timestamp_key,
    ).as_ref()

    qt = DeriveValidityPeriodNode(dialect, compute_mode, qt, fdw, effective_timestamp()).as_ref()

    qt = TrimValidityPeriodNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=qt,
        start=query_time_range.start,
        end=query_time_range.end,
    ).as_ref()

    return qt


def build_aggregated_time_range_ghf_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    from_source: Optional[bool],
    feature_data_time_limits: pendulum.Period,
    query_time_range: pendulum.Period,
    entities: Optional[DataframeWrapper] = None,
) -> NodeRef:
    partial_aggs = build_get_features(
        dialect,
        compute_mode,
        fdw,
        from_source,
        feature_data_time_limits=feature_data_time_limits,
    )
    spine = _build_internal_spine(dialect, compute_mode, fdw, partial_aggs)

    # TODO(danny): Don't represent the partial_agg node twice (in the spine + in the agg node)
    #  When this QT is compiled to SQL, the partial agg node is doubly represented and executed.
    full_aggregation_node = AsofJoinFullAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        spine=spine,
        partial_agg_node=partial_aggs,
        fdw=fdw,
        # Do not push down the timestamp if the spine is completely from partial agg, such as `ghf` and `run` with time
        # range.
        enable_spine_time_pushdown_rewrite=False,
        enable_spine_entity_pushdown_rewrite=False,
    ).as_ref()

    if fdw.aggregation_secondary_key:
        full_aggregation_node = AggregationSecondaryKeyRollupNode(
            dialect=dialect,
            compute_mode=compute_mode,
            full_aggregation_node=full_aggregation_node,
            fdw=fdw,
            group_by_columns=[*list(fdw.join_keys), anchor_time()],
        ).as_ref()

    qt = _filter_and_update_timestamp_columns(
        full_aggregation_node,
        dialect,
        compute_mode,
        fdw,
        respect_feature_start_time=True,
        show_effective_time=True,
    )

    if entities is not None:
        qt = _filter_entity_dataframe(dialect, compute_mode, qt, entities)

    qt = _filter_by_time_range(dialect, compute_mode, qt, query_time_range, fdw.timestamp_key, fdw.timestamp_key)

    return qt


def build_aggregated_time_range_run_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    feature_data_time_limits: pendulum.Period,
    aggregation_anchor_time: Optional[datetime] = None,
) -> NodeRef:
    partial_aggs = build_get_features(
        dialect,
        compute_mode,
        fdw,
        from_source=True,
        feature_data_time_limits=feature_data_time_limits,
        aggregation_anchor_time=aggregation_anchor_time,
    )
    spine = _build_internal_spine(dialect, compute_mode, fdw, partial_aggs)

    # TODO(danny): Don't represent the partial_agg node twice (in the spine + in the agg node)
    #  When this QT is compiled to SQL, the partial agg node is doubly represented and executed.
    full_aggregation_node = AsofJoinFullAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        spine=spine,
        partial_agg_node=partial_aggs,
        fdw=fdw,
        # Do not push down the timestamp if the spine is completely from partial agg, such as `ghf` and `run` with time
        # range.
        enable_spine_time_pushdown_rewrite=False,
        enable_spine_entity_pushdown_rewrite=False,
    ).as_ref()

    if fdw.aggregation_secondary_key:
        full_aggregation_node = AggregationSecondaryKeyRollupNode(
            dialect=dialect,
            compute_mode=compute_mode,
            full_aggregation_node=full_aggregation_node,
            fdw=fdw,
            group_by_columns=[*list(fdw.join_keys), anchor_time()],
        ).as_ref()
    qt = _filter_and_update_timestamp_columns(
        full_aggregation_node,
        dialect,
        compute_mode,
        fdw,
        respect_feature_start_time=False,
        show_effective_time=False,
    )

    return qt


def _filter_by_time_range(
    dialect: Dialect,
    compute_mode: ComputeMode,
    qt: NodeRef,
    time_range: pendulum.Period,
    start_timestamp_field: str,
    end_timestamp_field: str,
) -> NodeRef:
    qt = FeatureTimeFilterNode(
        dialect,
        compute_mode,
        qt,
        feature_data_time_limits=time_range,
        policy=feature_view__data_pb2.MaterializationTimeRangePolicy.MATERIALIZATION_TIME_RANGE_POLICY_FILTER_TO_RANGE,
        start_timestamp_field=start_timestamp_field,
        end_timestamp_field=end_timestamp_field,
    ).as_ref()

    return qt


def _filter_entity_dataframe(
    dialect: Dialect,
    compute_mode: ComputeMode,
    qt: NodeRef,
    entities: Optional[DataframeWrapper],
) -> NodeRef:
    columns = list(entities.columns)
    entities_df = SelectDistinctNode(
        dialect, compute_mode, UserSpecifiedDataNode(dialect, compute_mode, entities).as_ref(), columns
    ).as_ref()
    qt = JoinNode(dialect, compute_mode, qt, entities_df, columns, how="right").as_ref()
    return qt


def _filter_and_update_timestamp_columns(
    full_aggregation_node: NodeRef,
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    respect_feature_start_time: bool = True,
    show_effective_time: bool = False,
) -> NodeRef:
    """
    1. Filters rows by Feature Start Time
    2. Converts the anchor time to a timestamp from epoch ns.
    3. Adjusts the anchor time to reflect the end of the materialization window
    4. Optionally adds the _effective_time column
    """

    if respect_feature_start_time and fdw.feature_start_timestamp:
        full_aggregation_node = RespectFeatureStartTimeNode.for_anchor_time_column(
            dialect,
            compute_mode,
            full_aggregation_node,
            anchor_time(),
            fdw,
        ).as_ref()

    # The `AsofJoinFullAggNode` returned by `build_get_full_agg_features` converts timestamps to epochs. We convert back
    # from epochs to timestamps so that we can add an effective timestamp column.
    qt = ConvertEpochToTimestampNode(
        dialect, compute_mode, full_aggregation_node, {anchor_time(): fdw.get_feature_store_format_version}
    ).as_ref()

    # We want the time to be on the end of the window not the start.
    qt = AddDurationNode(
        dialect,
        compute_mode,
        qt,
        timestamp_field=anchor_time(),
        duration=fdw.get_tile_interval_duration_for_offline_store,
        new_column_name=fdw.trailing_time_window_aggregation().time_key,
    ).as_ref()
    qt = RenameColsNode(dialect, compute_mode, qt, drop=[anchor_time()]).as_ref()

    if show_effective_time:
        batch_schedule_seconds = 0 if fdw.is_feature_table else fdw.batch_materialization_schedule.in_seconds()
        qt = AddEffectiveTimestampNode(
            dialect,
            compute_mode,
            qt,
            timestamp_field=fdw.trailing_time_window_aggregation().time_key,
            effective_timestamp_name=effective_timestamp(),
            batch_schedule_seconds=batch_schedule_seconds,
            data_delay_seconds=fdw.online_store_data_delay_seconds,
            is_stream=fdw.is_stream,
            is_temporal_aggregate=True,
        ).as_ref()

    return qt


def _build_internal_spine(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    partial_aggs: NodeRef,
    explode_anchor_time: Optional[bool] = False,
    query_time_range: Optional[pendulum.Period] = None,
) -> NodeRef:
    if fdw.aggregation_secondary_key:
        cols_to_drop = list(
            set(partial_aggs.columns) - {*list(fdw.join_keys), anchor_time(), fdw.aggregation_secondary_key}
        )
    else:
        cols_to_drop = list(set(partial_aggs.columns) - {*list(fdw.join_keys), anchor_time()})

    spine = RenameColsNode(dialect, compute_mode, partial_aggs, drop=cols_to_drop).as_ref()

    # TODO (ajeya): remove flag after GHF is deprecated
    if explode_anchor_time:
        earliest_valid_anchor_time = time_utils.get_nearest_anchor_time(
            timestamp=query_time_range.start,
            max_source_data_delay=fdw.max_source_data_delay,
            batch_materialization_schedule=fdw.batch_materialization_schedule,
            min_scheduling_interval=fdw.min_scheduling_interval,
        )
        spine_filter = pendulum.Period(earliest_valid_anchor_time, query_time_range.end)
        spine = ExplodeTimestampByTimeWindowsNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=spine,
            timestamp_field=anchor_time(),
            time_filter=spine_filter,
            fdw=fdw,
        ).as_ref()

    if fdw.aggregation_secondary_key:
        spine = AggregationSecondaryKeyExplodeNode.for_feature_definition(
            dialect=dialect, compute_mode=compute_mode, spine=spine, fdw=fdw
        )

    return spine


def build_spine_join_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    dac: FeatureDefinitionAndJoinConfig,
    spine_node: NodeRef,
    spine_time_field: str,
    from_source: Optional[bool],
    use_namespace_feature_prefix: bool = True,
) -> NodeRef:
    fdw = dac.feature_definition
    if fdw.timestamp_key is not None and spine_time_field != fdw.timestamp_key:
        spine_node = RenameColsNode(
            dialect, compute_mode, spine_node, mapping={spine_time_field: fdw.timestamp_key}
        ).as_ref()
    if any(jk[0] != jk[1] for jk in dac.join_keys):
        spine_node = RenameColsNode(
            dialect, compute_mode, spine_node, mapping={jk[0]: jk[1] for jk in dac.join_keys if jk[0] != jk[1]}
        ).as_ref()
    if fdw.is_temporal or fdw.is_feature_table:
        ret = _build_spine_query_tree_temporal_or_feature_table(
            dialect=dialect,
            compute_mode=compute_mode,
            spine_node=spine_node,
            dac=dac,
            data_delay_seconds=fdw.online_store_data_delay_seconds,
            from_source=from_source,
            use_namespace_feature_prefix=use_namespace_feature_prefix,
        )
    elif fdw.is_temporal_aggregate:
        anchor_time_column = None
        columns_to_drop = None
        if fdw.get_min_batch_sawtooth_tile_size():
            sawtooth_agg_data = compaction_utils.SawtoothAggregationData(
                batch_sawtooth_tile_sizes=fdw.get_batch_sawtooth_tile_sizes()
            )
            full_agg_node = _build_sawtooth_aggregation_subtree(
                dialect=dialect,
                compute_mode=compute_mode,
                from_source=from_source,
                spine_node=spine_node,
                fdw=fdw,
                sawtooth_aggregation_data=sawtooth_agg_data,
            )
            anchor_time_column = sawtooth_agg_data.get_anchor_time_columns()[
                0
            ]  # Each anchor time column contains the same value for the spine.
            columns_to_drop = (
                sawtooth_agg_data.get_anchor_time_columns() + sawtooth_agg_data.get_identifier_partition_columns()
            )
        else:
            anchor_time_column = anchor_time()
            columns_to_drop = [anchor_time()]
            partial_agg_node = build_get_features(
                dialect,
                compute_mode,
                fdw,
                from_source=from_source,
                # NOTE: feature_data_time_limits is set to None since time pushdown
                # should happen as part of a optimization rewrite.
                feature_data_time_limits=None,
                aggregation_anchor_time=None,
            )
            augmented_spine = _augment_spine_for_window_aggregation(
                dialect,
                compute_mode,
                fdw,
                spine_node,
                partial_agg_node,
            )
            full_agg_node = AsofJoinFullAggNode(
                dialect=dialect,
                compute_mode=compute_mode,
                spine=augmented_spine,
                partial_agg_node=partial_agg_node,
                fdw=fdw,
                # Allow timestamp push down if the spine is provided by users.
                enable_spine_time_pushdown_rewrite=True,
                enable_spine_entity_pushdown_rewrite=True,
            ).as_ref()

        if fdw.aggregation_secondary_key:
            # Beside join keys and anchor time, we need to group by timestamp_key and TECTON_UNIQUE_ID_COL because:
            #   1. Grouping by timestamp_key is required to keep the timestamp column in the result.
            #   2. Grouping by TECTON_UNIQUE_ID_COL can distinguish duplicated rows in the spine.
            group_by_columns = [*list(fdw.join_keys), anchor_time(), fdw.timestamp_key, tecton_unique_id_col()]
            if tecton_spine_row_id_col() in full_agg_node.columns:
                group_by_columns += [tecton_spine_row_id_col()]
            full_agg_node = AggregationSecondaryKeyRollupNode(
                dialect=dialect,
                compute_mode=compute_mode,
                full_aggregation_node=full_agg_node,
                fdw=fdw,
                group_by_columns=group_by_columns,
            ).as_ref()
            full_agg_node = RenameColsNode(dialect, compute_mode, full_agg_node, drop=[tecton_unique_id_col()]).as_ref()

        if fdw.feature_start_timestamp:
            full_agg_node = RespectFeatureStartTimeNode.for_anchor_time_column(
                dialect,
                compute_mode,
                full_agg_node,
                anchor_time_column,
                fdw,
            ).as_ref()
        ret = _rename_feature_columns_and_drop_non_feature_columns(
            dialect,
            compute_mode,
            dac,
            full_agg_node,
            columns_to_drop,
            use_namespace_feature_prefix,
        )
    elif fdw.is_realtime:
        inputs = find_dependent_feature_set_items(
            fdw.fco_container,
            fdw.pipeline.root,
            visited_inputs={},
            fv_id=fdw.id,
        )
        dac = FeatureDefinitionAndJoinConfig.from_feature_definition(fdw)
        fsc = FeatureSetConfig([*inputs, dac])
        ret = build_feature_set_config_querytree(
            dialect, compute_mode, fsc, spine_node, spine_time_field, from_source, use_namespace_feature_prefix
        )
    else:
        raise NotImplementedError
    if fdw.timestamp_key is not None and spine_time_field != fdw.timestamp_key:
        ret = RenameColsNode(dialect, compute_mode, ret, {fdw.timestamp_key: spine_time_field}).as_ref()
    if any(jk[0] != jk[1] for jk in dac.join_keys):
        ret = RenameColsNode(
            dialect, compute_mode, ret, {jk[1]: jk[0] for jk in dac.join_keys if jk[0] != jk[1]}
        ).as_ref()
    return ret


def _update_internal_cols(
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    dac: FeatureDefinitionAndJoinConfig,
    internal_cols: Set[str],
) -> None:
    if dac.namespace.startswith(udf_internal()):
        for feature in fdw.features:
            internal_cols.add(dac.namespace + fdw.namespace_separator + feature)
    for feature in dac.features:
        if udf_internal() in feature:
            internal_cols.add(feature)


# Construct each wildcard materialized fvtree by joining against distinct set of join keys.
# Then, outer join these using WildcardJoinNode which performs an outer join while handling null-valued features properly.
def _build_wild_fv_subtree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    spine_node: NodeRef,
    fv_dacs: List[FeatureDefinitionAndJoinConfig],
    spine_time_field: str,
    from_source: Optional[bool],
) -> NodeRef:
    newtree = None
    for dac in fv_dacs:
        fdw = dac.feature_definition

        subspine_join_keys = [jk[0] for jk in dac.join_keys if jk[0] != fdw.wildcard_join_key]
        # SelectDistinctNode is needed for correctness in order to filter out rows with duplicate join keys before
        # retrieving feature values. This avoids exploding wildcard rows when there are duplicates in both the spine and the
        # feature view tree.
        subspine = SelectDistinctNode(
            dialect, compute_mode, spine_node, [*subspine_join_keys, spine_time_field]
        ).as_ref()
        fvtree = build_spine_join_querytree(dialect, compute_mode, dac, subspine, spine_time_field, from_source)
        if len(dac.features) < len(fdw.features):
            fvtree = RenameColsNode(
                dialect,
                compute_mode,
                fvtree,
                drop=[f"{fdw.name}{fdw.namespace_separator}{f}" for f in fdw.features if f not in dac.features],
            ).as_ref()
        if newtree is None:
            newtree = fvtree
        else:
            join_cols = [*subspine_join_keys, spine_time_field, fdw.wildcard_join_key]
            newtree = WildcardJoinNode(dialect, compute_mode, newtree, fvtree, join_cols=join_cols).as_ref()
    return newtree


# Construct each non-wildcard materialized fvtree by joining against distinct set of join keys.
# Then, outer join these fvtrees together.
def _build_standard_fv_subtree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    spine_node: NodeRef,
    fv_dacs: List[FeatureDefinitionAndJoinConfig],
    spine_time_field: str,
    from_source: Optional[bool],
) -> Tuple[NodeRef, Set[str]]:
    internal_cols: Set[str] = set()

    if dialect == Dialect.DUCKDB:
        # unique (row) id column will be used as join key
        spine_node = AddUniqueIdNode(dialect, compute_mode, spine_node, column_name=tecton_spine_row_id_col()).as_ref()
        internal_cols.add(tecton_spine_row_id_col())

    newtree = spine_node
    for dac in fv_dacs:
        fdw = dac.feature_definition
        _update_internal_cols(fdw, dac, internal_cols)

        subspine_join_keys = [jk[0] for jk in dac.join_keys]
        if dialect != Dialect.DUCKDB:
            # SelectDistinctNode is needed for correctness in the case that there are duplicate rows in the spine. The
            # alternative considered was to add a row_id as a hash of the row or a monotonically increasing id, however the
            # row_id as a hash is not unique for duplicate rows and a monotonically increasing id is non-deterministic.
            # UPD: it doesn't apply to DuckDB dialect, which has a row id column now (see above)
            subspine = SelectDistinctNode(
                dialect, compute_mode, spine_node, [*subspine_join_keys, spine_time_field]
            ).as_ref()
        else:
            # keep only join keys, timestamp, row id
            subspine = RenameColsNode(
                dialect,
                compute_mode,
                spine_node,
                drop=set(spine_node.columns) - {*subspine_join_keys, spine_time_field, tecton_spine_row_id_col()},
            ).as_ref()

        fvtree = build_spine_join_querytree(dialect, compute_mode, dac, subspine, spine_time_field, from_source)
        if len(dac.features) < len(fdw.features):
            fvtree = RenameColsNode(
                dialect,
                compute_mode,
                fvtree,
                drop=[f"{fdw.name}{fdw.namespace_separator}{f}" for f in fdw.features if f not in dac.features],
            ).as_ref()
        if dialect == Dialect.DUCKDB:
            # Since we're joining on row id, we don't need these columns anymore
            fvtree = RenameColsNode(
                dialect,
                compute_mode,
                fvtree,
                drop=[*subspine_join_keys, spine_time_field],
            ).as_ref()

        allow_null_features = conf.get_bool("ALLOW_NULL_FEATURES")
        newtree = JoinNode(
            dialect,
            compute_mode,
            newtree,
            fvtree,
            how="inner",
            join_cols=(
                [*subspine_join_keys, spine_time_field] if dialect != Dialect.DUCKDB else [tecton_spine_row_id_col()]
            ),
            allow_nulls=allow_null_features,
        ).as_ref()
    return newtree, internal_cols


# Compute odfvs via udf on the parent (not using joins)
def _build_odfv_subtree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    parent_tree: NodeRef,
    odfv_dacs: List[FeatureDefinitionAndJoinConfig],
    spine_timestamp_field: str,
    use_namespace_feature_prefix: bool = True,
) -> NodeRef:
    newtree = StagingNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=parent_tree,
        staging_table_name=odfv_internal_staging_table(),
        query_tree_step=QueryTreeStep.AGGREGATION,
    ).as_ref()
    feature_definitions_namespaces = [(dac.feature_definition, dac.namespace) for dac in odfv_dacs]
    newtree = MultiOdfvPipelineNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=newtree,
        feature_definition_namespaces=feature_definitions_namespaces,
        events_df_timestamp_field=spine_timestamp_field,
        use_namespace_feature_prefix=use_namespace_feature_prefix,
    ).as_ref()

    # Compute the union of the features to be computed
    dac_features = set()
    fdw_features = set()
    for dac in odfv_dacs:
        feature_prefix = f"{dac.namespace}{dac.feature_definition.namespace_separator}"
        dac_features.update({f"{feature_prefix}{f}" for f in dac.features})
        fdw_features.update({f"{feature_prefix}{f}" for f in dac.feature_definition.features})

    # Drop features if user queried a subset via feature services
    if len(dac_features) < len(fdw_features):
        newtree = RenameColsNode(
            dialect,
            compute_mode,
            newtree,
            drop=[namespaced_feat for namespaced_feat in fdw_features if namespaced_feat not in dac_features],
        ).as_ref()
    return newtree


# Construct each materialized fvtree by joining against distinct set of join keys.
# Then, join the full spine against each of those.
# Finally, compute odfvs via udf on top of the result (not using joins)
def build_feature_set_config_querytree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fsc: FeatureSetConfig,
    spine_node: NodeRef,
    spine_time_field: str,
    from_source: Optional[bool],
    use_namespace_feature_prefix: bool = True,
) -> NodeRef:
    odfv_dacs: List[FeatureDefinitionAndJoinConfig] = []
    wildcard_dacs: List[FeatureDefinitionAndJoinConfig] = []
    normal_fv_dacs: List[FeatureDefinitionAndJoinConfig] = []

    for dac in fsc.definitions_and_configs:
        if dac.feature_definition.is_realtime:
            odfv_dacs.append(dac)
        elif _is_wildcard_join(dac.feature_definition, spine_node):
            wildcard_dacs.append(dac)
        else:
            normal_fv_dacs.append(dac)

    if wildcard_dacs:
        newtree = _build_wild_fv_subtree(
            dialect, compute_mode, spine_node, wildcard_dacs, spine_time_field, from_source
        )
    else:
        newtree = spine_node

    internal_cols: Set[str] = set()
    if normal_fv_dacs:
        newtree, internal_cols = _build_standard_fv_subtree(
            dialect, compute_mode, newtree, normal_fv_dacs, spine_time_field, from_source
        )

    if odfv_dacs:
        newtree = _build_odfv_subtree(
            dialect, compute_mode, newtree, odfv_dacs, spine_time_field, use_namespace_feature_prefix
        )

    # drop all internal cols
    if len(internal_cols) > 0:
        newtree = RenameColsNode(dialect, compute_mode, newtree, drop=list(internal_cols)).as_ref()

    return newtree


def _build_spine_query_tree_temporal_or_feature_table(
    dialect: Dialect,
    compute_mode: ComputeMode,
    spine_node: NodeRef,
    dac: FeatureDefinitionAndJoinConfig,
    data_delay_seconds: int,
    from_source: Optional[bool],
    use_namespace_feature_prefix: bool = True,
) -> NodeRef:
    fdw = dac.feature_definition
    base = build_get_features(dialect, compute_mode, fdw, from_source=from_source)
    batch_schedule_seconds = 0 if fdw.is_feature_table else fdw.batch_materialization_schedule.in_seconds()
    base = AddEffectiveTimestampNode(
        dialect,
        compute_mode,
        base,
        timestamp_field=fdw.timestamp_key,
        effective_timestamp_name=effective_timestamp(),
        batch_schedule_seconds=batch_schedule_seconds,
        data_delay_seconds=data_delay_seconds,
        is_stream=fdw.is_stream,
        is_temporal_aggregate=False,
    ).as_ref()
    if fdw.serving_ttl is not None:
        base = AddDurationNode(
            dialect,
            compute_mode,
            base,
            timestamp_field=fdw.timestamp_key,
            duration=fdw.serving_ttl,
            new_column_name=timestamp_plus_ttl(),
        ).as_ref()
        # Calculate effective expiration time = window(feature_time + ttl, batch_schedule).end + data_delay
        batch_schedule_seconds = 0 if fdw.is_feature_table else fdw.batch_materialization_schedule.in_seconds()
        base = AddEffectiveTimestampNode(
            dialect,
            compute_mode,
            base,
            timestamp_field=timestamp_plus_ttl(),
            effective_timestamp_name=expiration_timestamp(),
            batch_schedule_seconds=batch_schedule_seconds,
            data_delay_seconds=data_delay_seconds,
            is_stream=fdw.is_stream,
            is_temporal_aggregate=False,
        ).as_ref()
    rightside_join_prefix = default_case("_tecton_right")
    join_prefixed_feature_names = [f"{rightside_join_prefix}_{f}" for f in fdw.features]
    prefix = f"{dac.namespace}{fdw.namespace_separator}" if use_namespace_feature_prefix else ""
    # we can't just ask for the correct right_prefix to begin with because the asofJoin always sticks an extra underscore in between
    rename_map: Dict[str, Optional[str]] = {}
    cols_to_drop = []
    for f in fdw.features:
        if f not in dac.features:
            cols_to_drop.append(f"{rightside_join_prefix}_{f}")
        else:
            rename_map[f"{rightside_join_prefix}_{f}"] = f"{prefix}{f}"

    expiration_timestamp_col = f"{rightside_join_prefix}_{expiration_timestamp()}"

    cols_to_drop.append(f"{rightside_join_prefix}_{fdw.timestamp_key}")
    cols_to_drop.append(f"{rightside_join_prefix}_{anchor_time()}")
    cols_to_drop.append(f"{rightside_join_prefix}_{effective_timestamp()}")
    if fdw.serving_ttl is not None:
        cols_to_drop.append(f"{rightside_join_prefix}_{timestamp_plus_ttl()}")
        cols_to_drop.append(expiration_timestamp_col)

    if fdw.feature_start_timestamp is not None:
        base = RespectFeatureStartTimeNode(
            dialect,
            compute_mode,
            base,
            fdw.timestamp_key,
            fdw.feature_start_timestamp,
            fdw.features,
            fdw.get_feature_store_format_version,
        ).as_ref()

    if _is_wildcard_join(fdw, spine_node):
        # Need to copy base so that the left and right side are separate
        base_copy = base.deepcopy()
        spine_node = AsofSecondaryKeyExplodeNode(
            dialect, compute_mode, spine_node, fdw.timestamp_key, base_copy, effective_timestamp(), fdw
        ).as_ref()

    base = AsofJoinNode(
        dialect=dialect,
        compute_mode=compute_mode,
        left_container=AsofJoinInputContainer(spine_node, fdw.timestamp_key),
        right_container=AsofJoinInputContainer(
            base,
            timestamp_field=fdw.timestamp_key,
            effective_timestamp_field=effective_timestamp(),
            prefix=rightside_join_prefix,
            schema=fdw.view_schema,
        ),
        join_cols=fdw.join_keys,
    ).as_ref()

    if fdw.serving_ttl is not None:
        base = RespectTTLNode(
            dialect, compute_mode, base, fdw.timestamp_key, expiration_timestamp_col, join_prefixed_feature_names
        ).as_ref()
    # remove anchor cols/dupe timestamp cols
    return RenameColsNode(dialect, compute_mode, base, mapping=rename_map, drop=cols_to_drop).as_ref()


def _augment_spine_for_window_aggregation(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    spine_node: NodeRef,
    partial_agg_node: NodeRef,
) -> NodeRef:
    augmented_spine = AddRetrievalAnchorTimeNode(
        dialect,
        compute_mode,
        spine_node,
        name=fdw.name,
        feature_store_format_version=fdw.get_feature_store_format_version,
        batch_schedule=fdw.get_batch_schedule_for_version,
        tile_interval=fdw.get_tile_interval_for_offline_store,
        timestamp_field=fdw.timestamp_key,
        is_stream=fdw.is_stream,
        data_delay_seconds=fdw.online_store_data_delay_seconds,
    ).as_ref()

    # We need to explode the spine for the secondary key if:
    #     1. A Feature View with an aggregation_secondary_key: an aggregation_secondary_key is never in the spine,
    #     so we always need to explode the spine for it.
    #     2. A Feature View with a wild card join key: a wild card join key is optional in the spine, so we need to
    #     check if the wild card join key is not in the spine before exploding the spine for it.
    if fdw.aggregation_secondary_key or _is_wildcard_join(fdw, spine_node):
        # Add a unique id column if aggreagtion secondary key appears. The unique id column is used to make each spine
        # row unique so later the secondary key aggregation rollup doesn't merge duplicate rows.
        if fdw.aggregation_secondary_key:
            augmented_spine = AddUniqueIdNode(dialect, compute_mode, augmented_spine).as_ref()

        return AsofSecondaryKeyExplodeNode(
            dialect,
            compute_mode,
            augmented_spine,
            anchor_time(),
            partial_agg_node,
            anchor_time(),
            fdw,
        ).as_ref()

    return augmented_spine


def _rename_feature_columns_and_drop_non_feature_columns(
    dialect: Dialect,
    compute_mode: ComputeMode,
    dac: FeatureDefinitionAndJoinConfig,
    node: NodeRef,
    columns_to_drop: List[str],
    use_namespace_feature_prefix: bool = True,
) -> NodeRef:
    rename_map: Dict[str, Optional[str]] = {}
    for f in dac.feature_definition.features:
        if f not in dac.features:
            columns_to_drop.append(f)
        elif use_namespace_feature_prefix:
            # TODO: make a helper
            rename_map[f] = f"{dac.namespace}{dac.feature_definition.namespace_separator}{f}"
    return RenameColsNode(dialect, compute_mode, node, mapping=rename_map, drop=columns_to_drop).as_ref()


def _add_sawtooth_anchor_time_and_partition_columns(
    dialect: Dialect,
    compute_mode: ComputeMode,
    input_node: NodeRef,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    sawtooth_aggregation_data: compaction_utils.SawtoothAggregationData,
    timestamp_field: str,
    should_truncate_timestamps: bool,
) -> NodeRef:
    output_node = AddAnchorTimeColumnsForSawtoothIntervalsNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=input_node,
        timestamp_field=timestamp_field,
        anchor_time_column_map=sawtooth_aggregation_data.get_anchor_time_to_timedelta_map(
            use_zero_timedelta=not should_truncate_timestamps
        ),
        data_delay_seconds=fdw.online_store_data_delay_seconds,
        feature_store_format_version=fdw.get_feature_store_format_version,
        aggregation_tile_interval=fdw.get_tile_interval_for_offline_store,
    ).as_ref()
    partition_columns = sawtooth_aggregation_data.get_identifier_partition_columns()
    partial_column_to_bool_map = {column: True for column in partition_columns}
    output_node = AddBooleanPartitionColumnsNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=output_node,
        column_to_bool_map=partial_column_to_bool_map,
    ).as_ref()
    return output_node


def _build_sawtooth_aggregation_subtree(
    dialect: Dialect,
    compute_mode: ComputeMode,
    from_source: bool,
    spine_node: NodeRef,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    sawtooth_aggregation_data: compaction_utils.SawtoothAggregationData,
) -> NodeRef:
    """Compute the full agg rollups for aggregate fvs that use sawtoothing.

    Only applicable to stream fvs that use compaction."""

    # Build partial aggregates
    partial_agg_node = build_get_features(
        dialect,
        compute_mode,
        fdw,
        from_source=from_source,
        # NOTE: feature_data_time_limits is set to None since time pushdown
        # should happen as part of a optimization rewrite.
        feature_data_time_limits=None,
        aggregation_anchor_time=None,
    )
    partial_agg_node = _add_sawtooth_anchor_time_and_partition_columns(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=partial_agg_node,
        sawtooth_aggregation_data=sawtooth_aggregation_data,
        fdw=fdw,
        timestamp_field=anchor_time(),
        should_truncate_timestamps=False,
    )

    # Build spine for batch part full aggs.
    # TODO(samantha): Fix the spine creation for secondary key aggs.
    batch_spine_node = AddAnchorTimeColumnsForSawtoothIntervalsNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=spine_node,
        timestamp_field=fdw.timestamp_key,
        anchor_time_column_map=sawtooth_aggregation_data.get_anchor_time_to_timedelta_map(),
        data_delay_seconds=fdw.online_store_data_delay_seconds,
        feature_store_format_version=fdw.get_feature_store_format_version,
        aggregation_tile_interval=fdw.get_tile_interval_for_offline_store,
    ).as_ref()
    batch_spine_node = ExplodeEventsAndSelectDistinctNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=batch_spine_node,
        explode_columns=sawtooth_aggregation_data.get_anchor_time_columns(),
        explode_columns_to_boolean_columns=sawtooth_aggregation_data.get_anchor_time_to_partition_columns_map(),
    ).as_ref()

    batch_part_subtree = AsofJoinFullAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        spine=batch_spine_node,
        partial_agg_node=partial_agg_node,
        fdw=fdw,
        # Allow timestamp push down if the spine is provided by users.
        enable_spine_time_pushdown_rewrite=True,
        enable_spine_entity_pushdown_rewrite=True,
        sawtooth_aggregation_data=sawtooth_aggregation_data,
    ).as_ref()
    # The result of the full agg node has an anchor time that's the start of the day/hour.
    # For the AsofJoinSawtoothAggNode, this anchor time needs to be the end of the day/hour in order for the time partitioning to work correctly.
    batch_part_subtree = AdjustAnchorTimeToWindowEndNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=batch_part_subtree,
        anchor_time_columns=sawtooth_aggregation_data.get_anchor_time_columns(),
        aggregation_tile_interval=fdw.get_tile_interval_for_offline_store,
    ).as_ref()

    # Stream partial aggs
    stream_events_tree = _build_get_features(
        dialect=dialect,
        compute_mode=compute_mode,
        fdw=fdw,
        from_source=from_source,
        feature_data_time_limits=None,
        aggregation_anchor_time=None,
    )
    stream_partial_agg_tree = PartialAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=stream_events_tree,
        fdw=fdw,
        window_start_column_name=anchor_time(),
        aggregation_tile_interval=pendulum.Duration(
            seconds=0
        ),  # Explicity set 0 because we need the raw offline events.
        aggregation_anchor_time=None,
    ).as_ref()
    # TODO(samantha): Add pushdown filtering for entities for the stream partial aggs
    stream_partial_agg_tree = _add_sawtooth_anchor_time_and_partition_columns(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=stream_partial_agg_tree,
        sawtooth_aggregation_data=sawtooth_aggregation_data,
        fdw=fdw,
        timestamp_field=anchor_time(),
        should_truncate_timestamps=False,
    )

    sawtooth_agg_spine_tree = _add_sawtooth_anchor_time_and_partition_columns(
        dialect=dialect,
        compute_mode=compute_mode,
        input_node=spine_node,
        sawtooth_aggregation_data=sawtooth_aggregation_data,
        fdw=fdw,
        timestamp_field=fdw.timestamp_key,
        should_truncate_timestamps=False,
    )
    # Build final aggregation values. This is a placeholder node for now.
    sawtooth_agg_node = AsofJoinSawtoothAggNode(
        dialect=dialect,
        compute_mode=compute_mode,
        batch_input_node=batch_part_subtree,
        stream_input_node=stream_partial_agg_tree,
        spine_input_node=sawtooth_agg_spine_tree,
        sawtooth_aggregation_data=sawtooth_aggregation_data,
        fdw=fdw,
    ).as_ref()
    return sawtooth_agg_node


def _build_aggregation_group_data_node(
    dialect: Dialect, compute_mode: ComputeMode, aggregation_groups: Tuple[compaction_utils.AggregationGroup, ...]
) -> PythonDataNode:
    data = tuple(
        (group.window_index, group.tile_index, group.inclusive_start_time, group.exclusive_end_time)
        for group in aggregation_groups
    )
    # We need to specify an explict schema since inclusive_start_time can be null.
    schema_dict = {
        aggregation_group_id(): data_types.Int64Type(),
        aggregation_tile_id(): data_types.Int64Type(),
        inclusive_start_time(): data_types.TimestampType(),
        exclusive_end_time(): data_types.TimestampType(),
    }
    return PythonDataNode.from_schema(dialect, compute_mode, schema=Schema.from_dict(schema_dict), data=data)


def build_compaction_query(
    dialect: Dialect,
    compute_mode: ComputeMode,
    fdw: feature_definition_wrapper.FeatureDefinitionWrapper,
    compaction_job_end_time: datetime,
    enable_from_source_for_test: bool = False,
) -> NodeRef:
    """Build compaction query for online materialization jobs. Only used for fvs with batch compaction enabled.

    enable_from_source_for_test=True should only be used in local integration tests!"""
    feature_data_time_limits = compaction_utils.get_data_time_limits_for_compaction(
        fdw=fdw, compaction_job_end_time=compaction_job_end_time
    )
    if enable_from_source_for_test:
        # Only used for local testing
        base_node = build_pipeline_querytree(
            dialect, compute_mode, fdw, for_stream=False, feature_data_time_limits=feature_data_time_limits
        )
    else:
        base_node = OfflineStoreScanNode(
            dialect=dialect,
            compute_mode=compute_mode,
            feature_definition_wrapper=fdw,
            partition_time_filter=feature_data_time_limits,
        ).as_ref()
        if feature_data_time_limits:
            base_node = FeatureTimeFilterNode(
                dialect=dialect,
                compute_mode=compute_mode,
                input_node=base_node,
                feature_data_time_limits=feature_data_time_limits,
                policy=fdw.time_range_policy,
                start_timestamp_field=fdw.timestamp_key,
                end_timestamp_field=fdw.timestamp_key,
            ).as_ref()

    if fdw.is_temporal_aggregate:
        aggregation_groups = compaction_utils.aggregation_groups(fdw, compaction_job_end_time)
        compaction_ranges = _build_aggregation_group_data_node(dialect, compute_mode, aggregation_groups).as_ref()
        node = InnerJoinOnRangeNode(
            dialect=dialect,
            compute_mode=compute_mode,
            left=base_node,
            right=compaction_ranges,
            left_join_condition_column=fdw.timestamp_key,
            right_inclusive_start_column=inclusive_start_time(),
            right_exclusive_end_column=exclusive_end_time(),
        ).as_ref()
        node = OnlinePartialAggNodeV2(
            dialect,
            compute_mode,
            node,
            fdw=fdw,
            aggregation_groups=aggregation_groups,
        ).as_ref()
        node = OnlineListAggNode(
            dialect=dialect,
            compute_mode=compute_mode,
            input_node=node,
            fdw=fdw,
            aggregation_groups=aggregation_groups,
        ).as_ref()
        return node
    elif fdw.is_temporal:
        node = TakeLastRowNode.for_feature_definition(
            dialect=dialect, compute_mode=compute_mode, fdw=fdw, input_node=base_node
        )
        node = TemporalBatchTableFormatNode.for_feature_definition(
            dialect=dialect, compute_mode=compute_mode, fdw=fdw, input_node=node
        )
        return node

    msg = "Unexpected FV type."
    raise Exception(msg)



import json
import sys

import click

from tecton._internals import metadata_service
from tecton.cli import printer
from tecton.cli.command import TectonGroup
from tecton_core.errors import TectonAPIValidationError
from tecton_proto.metadataservice.metadata_service__client_pb2 import CreateServiceAccountRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import DeleteServiceAccountRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetServiceAccountsRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import UpdateServiceAccountRequest


@click.command("service-account", cls=TectonGroup)
def service_account():
    """Manage Service Accounts."""


@service_account.command()
@click.option("-n", "--name", required=True, help="Name of the Service Account")
@click.option(
    "-d", "--description", default="", help="An optional, human readable description for this Service Account"
)
@click.option("--json-out", default=False, is_flag=True, help="Format Output as JSON")
def create(name, description, json_out):
    """Create a new Service Account."""
    try:
        request = CreateServiceAccountRequest(name=name, description=description)
        response = metadata_service.instance().CreateServiceAccount(request)
    except TectonAPIValidationError as e:
        printer.safe_print(f"Failed to create service account: {e}", file=sys.stderr)
        sys.exit(1)
    if json_out:
        service_account = {}
        service_account["api_key"] = response.api_key
        service_account["id"] = response.id
        printer.safe_print(json.dumps(service_account, indent=4))
    else:
        printer.safe_print("Save this API Key - you will not be able to get it again.")
        printer.safe_print(f"API Key:            {response.api_key}")
        printer.safe_print(f"Service Account ID: {response.id}")
        printer.safe_print("Use `tecton access-control assign-role` to assign roles to your new service account.")


@service_account.command()
@click.argument("id", required=True)
def delete(id):
    """Permanently delete a Service Account by its ID."""
    request = DeleteServiceAccountRequest(id=id)
    try:
        response = metadata_service.instance().DeleteServiceAccount(request)
    except TectonAPIValidationError as e:
        printer.safe_print(f"Failed to delete service account: {e}", file=sys.stderr)
        sys.exit(1)
    printer.safe_print("Successfully deleted Service Account")


@service_account.command()
@click.option("-n", "--name", help="Name of the Service Account")
@click.option("-d", "--description", help="An optional, human readable description for this Service Account")
@click.argument("id", required=True)
def update(id, name, description):
    """Update the name or description of a Service Account."""
    request = UpdateServiceAccountRequest(id=id)

    if name is None and description is None:
        msg = "Please mention the field to update using --name or --description."
        raise click.ClickException(msg)

    if name:
        request.name = name

    if description is not None:
        request.description = description

    try:
        response = metadata_service.instance().UpdateServiceAccount(request)
    except TectonAPIValidationError as e:
        printer.safe_print(f"Failed to Update Service Account: {e}", file=sys.stderr)
        sys.exit(1)

    printer.safe_print("Successfully updated Service Account")


@service_account.command()
@click.argument("id", required=True)
def activate(id):
    """Activate a Service Account by its ID."""
    request = UpdateServiceAccountRequest(id=id, is_active=True)
    try:
        response = metadata_service.instance().UpdateServiceAccount(request)
    except TectonAPIValidationError as e:
        printer.safe_print(f"Failed to Activate Service Account: {e}", file=sys.stderr)
        sys.exit(1)
    printer.safe_print("Successfully activated Service Account")


@service_account.command()
@click.argument("id", required=True)
def deactivate(id):
    """Deactivate a Service Account by its ID. This disables the Service Account but does not permanently delete it."""
    request = UpdateServiceAccountRequest(id=id, is_active=False)
    try:
        response = metadata_service.instance().UpdateServiceAccount(request)
    except TectonAPIValidationError as e:
        printer.safe_print(f"Failed to Deactivate Service Account: {e}", file=sys.stderr)
        sys.exit(1)
    printer.safe_print("Successfully deactivated Service Account")


@service_account.command()
@click.option("--json-out", default=False, is_flag=True, help="Format Output as JSON")
@click.option("-s", "--search-string", help="Search String to search by ID, Name or Description")
def list(json_out, search_string):
    """List Service Accounts."""
    request = GetServiceAccountsRequest()

    if search_string is not None:
        request.search = search_string

    response = metadata_service.instance().GetServiceAccounts(request)
    service_accounts = []

    if len(response.service_accounts) == 0:
        printer.safe_print("No Service Accounts Found")
        return

    for k in response.service_accounts:
        if json_out:
            account = {}
            account["name"] = k.name
            account["id"] = k.id
            account["description"] = k.description
            account["active"] = k.is_active
            account["createdBy"] = k.created_by
            service_accounts.append(account)
        else:
            printer.safe_print(f"{'Name: ': <15}{k.name}")
            printer.safe_print(f"{'ID: ': <15}{k.id}")
            if k.description:
                printer.safe_print(f"{'Description: ': <15}{k.description}")
            if k.created_by:
                if k.created_by.HasField("user"):
                    printer.safe_print(f"{'Created By: ': <15}{k.created_by.user.login_email : <20}(User Email)")
                elif k.created_by.HasField("service_account"):
                    printer.safe_print(
                        f"{'Created By: ': <15}{k.created_by.service_account.name : <20}(Service Account Name)"
                    )
            printer.safe_print(f"{'Active: ': <15}{k.is_active}")
            printer.safe_print()
    if json_out:
        printer.safe_print(json.dumps(service_accounts, indent=4))



"""Utilities for deriving data source and feature view schemas. Shared by backend and local schema derivation."""

import datetime
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Sequence

import pendulum
import pyspark
from pyspark.sql import types as pyspark_types
from typeguard import typechecked

from tecton_core import conf
from tecton_core import errors as core_errors
from tecton_core import filter_context
from tecton_core import specs
from tecton_core.id_helper import IdHelper
from tecton_proto.args import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as virtual_data_source__args_pb2
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.common import spark_schema__client_pb2 as spark_schema_pb2
from tecton_proto.common.data_source_type__client_pb2 import DataSourceType
from tecton_spark import data_source_helper
from tecton_spark import errors_spark
from tecton_spark import schema_spark_utils
from tecton_spark import spark_schema_wrapper
from tecton_spark.spark_pipeline import SparkFeaturePipeline
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper
from tecton_spark.type_annotations import PySparkDataFrame
from tecton_spark.type_annotations import PySparkSession


@typechecked
def get_hive_table_schema(
    spark: PySparkSession,
    database: str,
    table: str,
    post_processor: Optional[Callable],
    timestamp_field: str,
    timestamp_format: str,
) -> spark_schema_pb2.SparkSchema:
    df = data_source_helper._get_raw_hive_table_dataframe(spark, database, table)
    if post_processor is not None:
        df = post_processor(df)
    if timestamp_field:
        ts_format = None
        if timestamp_format:
            ts_format = timestamp_format
        df = data_source_helper.apply_timestamp_column(df, timestamp_field, ts_format)
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


def get_unity_table_schema(
    spark: PySparkSession,
    catalog: str,
    schema: str,
    table: str,
    post_processor: Optional[Callable],
    timestamp_field: str,
    timestamp_format: str,
) -> spark_schema_pb2.SparkSchema:
    df = data_source_helper._get_raw_unity_table_dataframe(spark, catalog, schema, table)
    if post_processor is not None:
        df = post_processor(df)
    if timestamp_field:
        ts_format = None
        if timestamp_format:
            ts_format = timestamp_format
        df = data_source_helper.apply_timestamp_column(df, timestamp_field, ts_format)
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_redshift_table_schema(
    spark: PySparkSession,
    endpoint: str,
    table: str,
    query: str,
    temp_s3: str,
    post_processor: Optional[Callable],
) -> spark_schema_pb2.SparkSchema:
    df = data_source_helper.get_redshift_dataframe(spark, endpoint, temp_s3, table=table, query=query)
    if post_processor is not None:
        df = post_processor(df)
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_snowflake_schema(
    spark: PySparkSession,
    url: str,
    database: str,
    schema: str,
    warehouse: str,
    role: Optional[str],
    table: Optional[str],
    query: Optional[str],
    post_processor: Optional[Callable],
) -> spark_schema_pb2.SparkSchema:
    assert table is not None or query is not None, "Both table and query cannot be None"

    df = data_source_helper.get_snowflake_dataframe(
        spark,
        url,
        database,
        schema,
        warehouse,
        role=role,
        table=table,
        query=query,
    )
    if post_processor is not None:
        df = post_processor(df)
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_batch_data_source_function_schema(
    spark: PySparkSession, data_source_function: Callable, supports_time_filtering: bool
) -> spark_schema_pb2.SparkSchema:
    if supports_time_filtering:
        df_fc_none = data_source_function(spark=spark, filter_context=None)
        df_fc_none_start_none_end = data_source_function(
            spark=spark, filter_context=filter_context.FilterContext(None, None)
        )
        df_fc_none_end = data_source_function(
            spark=spark, filter_context=filter_context.FilterContext(pendulum.datetime(1970, 1, 1), None)
        )
        df_fc_none_start = data_source_function(
            spark=spark, filter_context=filter_context.FilterContext(None, pendulum.now())
        )
        schema = df_fc_none.schema
        # Verify filter_context is handled correctly. Schema should be the same for all values of filter_context.
        filter_context_error_message = (
            f"Invalid handling of filter_context and time filtering. Data Source Function {data_source_function.__name__} "
            f"needs to return a Spark DataFrame with the same schema for all values of filter_context"
        )
        assert all(
            df.schema == schema for df in [df_fc_none_start_none_end, df_fc_none_end, df_fc_none_start]
        ), filter_context_error_message

        df = df_fc_none
    else:
        df = data_source_function(spark=spark)
    assert isinstance(df, PySparkDataFrame), "Data Source Function must return a Spark DataFrame"
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


def get_file_source_schema(
    spark: PySparkSession,
    file_format: str,
    file_uri: str,
    convert_to_glue: bool,
    schema_uri: Optional[str],
    schema_override: Optional[spark_schema_wrapper.SparkSchemaWrapper],
    post_processor: Optional[Callable],
    timestamp_col: Optional[str],
    timestmap_format: Optional[str],
) -> spark_schema_pb2.SparkSchema:
    reader = spark.read
    if schema_uri is not None:
        uri = schema_uri
        assert schema_uri.startswith(file_uri), f"{schema_uri} must contain {file_uri}"
        # Setting basePath includes the path-based partitions in the DataFrame schema.
        # https://spark.apache.org/docs/latest/sql-data-sources-parquet.html#partition-discovery
        reader = reader.option("basePath", file_uri)
    else:
        uri = file_uri

    if schema_override is not None:
        reader = reader.schema(schema_override.unwrap())

    if file_format == "json":

        def action():
            return reader.json(uri)

    elif file_format == "parquet":

        def action():
            return reader.parquet(uri)

    elif file_format == "csv":

        def action():
            return reader.csv(uri, header=True)

    else:
        msg = f"Unsupported file format '{file_format}'"
        raise AssertionError(msg)

    df = errors_spark.handleDataAccessErrors(action, file_uri)

    if convert_to_glue:
        df = data_source_helper.convert_json_like_schema_to_glue_format(spark, df)
    if post_processor is not None:
        df = post_processor(df)

    if timestamp_col is not None:
        df = data_source_helper.apply_timestamp_column(df, timestamp_col, timestmap_format)

    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_kinesis_schema(
    spark: PySparkSession, stream_name: str, post_processor: Callable
) -> spark_schema_pb2.SparkSchema:
    """Compute the Kinesis schema using mock Kinesis data.

    Creates a mocked DataFrame for this stream, without actually creating a stream reader.
    This method returns a message in the Kinesis message format (below) with mocked contents.

    |-- approximateArrivalTimestamp: timestamp
    |-- data: binary
    |-- partitionKey: string
    |-- sequenceNumber: string
    |-- streamName: string
    """
    row = pyspark.Row(
        data=bytearray("no_data", "utf-8"),
        streamName=stream_name,
        partitionKey="0",
        sequenceNumber="0",
        approximateArrivalTimestamp=datetime.datetime.fromtimestamp(0),
    )
    df = spark.createDataFrame([row])

    df = post_processor(df)

    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


# https://docs.databricks.com/spark/latest/structured-streaming/kafka.html
KAFKA_SCHEMA = pyspark_types.StructType(
    [
        pyspark_types.StructField("key", pyspark_types.BinaryType(), True),
        pyspark_types.StructField("value", pyspark_types.BinaryType(), True),
        pyspark_types.StructField("topic", pyspark_types.StringType(), True),
        pyspark_types.StructField("partition", pyspark_types.IntegerType(), True),
        pyspark_types.StructField("offset", pyspark_types.LongType(), True),
        pyspark_types.StructField("timestamp", pyspark_types.TimestampType(), True),
        pyspark_types.StructField("timestampType", pyspark_types.IntegerType(), True),
    ]
)


@typechecked
def get_kafka_schema(spark: PySparkSession, post_processor: Callable) -> spark_schema_pb2.SparkSchema:
    """Compute the Kafka schema using mock Kafka data."""
    df = spark.createDataFrame([], KAFKA_SCHEMA)
    df = post_processor(df)
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_stream_data_source_function_schema(
    spark: PySparkSession, data_source_fn: Callable
) -> spark_schema_pb2.SparkSchema:
    """Compute the Kafka schema using mock Kafka data."""
    df = data_source_fn(spark=spark)
    assert (
        isinstance(df, PySparkDataFrame) and df.isStreaming
    ), "Data Source Function must return a streaming Spark DataFrame"
    return spark_schema_wrapper.SparkSchemaWrapper.from_spark_schema(df.schema)


@typechecked
def get_feature_view_view_schema(
    spark: PySparkSession,
    feature_view: feature_view_pb2.FeatureViewArgs,
    transformations: Sequence[specs.TransformationSpec],
    data_sources: Sequence[specs.DataSourceSpec],
) -> schema_pb2.Schema:
    """Compute the Feature View view schema."""
    has_push_source = False
    for data_source in data_sources:
        if data_source.type == DataSourceType.PUSH_WITH_BATCH or data_source.type == DataSourceType.PUSH_NO_BATCH:
            has_push_source = True
            break
    # This schema is only set for Stream Feature Views with PushSources.
    if feature_view.materialized_feature_view_args.schema and has_push_source and len(transformations) > 0:
        return feature_view.materialized_feature_view_args.schema
    df = get_feature_view_empty_view_df(spark, feature_view, transformations, data_sources)
    return schema_spark_utils.schema_from_spark(df.schema).to_proto()


@typechecked
def get_feature_view_empty_view_df(
    spark: PySparkSession,
    feature_view: feature_view_pb2.FeatureViewArgs,
    transformations: Sequence[specs.TransformationSpec],
    data_sources: Sequence[specs.DataSourceSpec],
) -> PySparkDataFrame:
    """Return a pyspark dataframe for the feature view "view" (i.e. before agggregations) using mock/empty data."""
    # Create empty data frames for each DS input matching the DS schema.
    id_to_ds = {ds.id: ds for ds in data_sources}
    empty_mock_inputs = populate_empty_passed_in_inputs(feature_view.pipeline.root, id_to_ds, spark)

    spark_pipeline = SparkFeaturePipeline(
        spark,
        pipeline=feature_view.pipeline,
        transformations=transformations,
        schedule_interval=_batch_schedule_from_fv(feature_view),
        data_source_inputs=empty_mock_inputs,
    )
    return spark_pipeline.to_dataframe()


def _batch_schedule_from_fv(feature_view: feature_view_pb2.FeatureViewArgs) -> Optional[pendulum.Duration]:
    if feature_view.HasField("materialized_feature_view_args"):
        return pendulum.Duration(seconds=feature_view.materialized_feature_view_args.batch_schedule.ToSeconds())
    else:
        return None


def derive_batch_schema(
    spark: PySparkSession,
    ds_args: virtual_data_source__args_pb2.VirtualDataSourceArgs,
    batch_post_processor: Optional[Callable],
    batch_data_source_function: Optional[Callable],
) -> spark_schema_pb2.SparkSchema:
    if ds_args.HasField("hive_ds_config"):
        return get_hive_table_schema(
            spark=spark,
            table=ds_args.hive_ds_config.table,
            database=ds_args.hive_ds_config.database,
            post_processor=batch_post_processor,
            timestamp_field=ds_args.hive_ds_config.common_args.timestamp_field,
            timestamp_format=ds_args.hive_ds_config.timestamp_format,
        )
    elif ds_args.HasField("unity_ds_config"):
        return get_unity_table_schema(
            spark=spark,
            catalog=ds_args.unity_ds_config.catalog,
            schema=ds_args.unity_ds_config.schema,
            table=ds_args.unity_ds_config.table,
            post_processor=batch_post_processor,
            timestamp_field=ds_args.unity_ds_config.common_args.timestamp_field,
            timestamp_format=ds_args.unity_ds_config.timestamp_format,
        )
    elif ds_args.HasField("spark_batch_config"):
        return get_batch_data_source_function_schema(
            spark=spark,
            data_source_function=batch_data_source_function,
            supports_time_filtering=ds_args.spark_batch_config.supports_time_filtering,
        )
    elif ds_args.HasField("redshift_ds_config"):
        if not ds_args.redshift_ds_config.HasField("endpoint"):
            msg = "redshift"
            raise core_errors.DS_ARGS_MISSING_FIELD(msg, "endpoint")

        has_table = ds_args.redshift_ds_config.HasField("table") and ds_args.redshift_ds_config.table
        has_query = ds_args.redshift_ds_config.HasField("query") and ds_args.redshift_ds_config.query
        if (has_table and has_query) or (not has_table and not has_query):
            raise core_errors.REDSHIFT_DS_EITHER_TABLE_OR_QUERY
        temp_s3 = conf.get_or_none("SPARK_REDSHIFT_TEMP_DIR")
        if temp_s3 is None:
            raise core_errors.REDSHIFT_DS_MISSING_SPARK_TEMP_DIR

        return get_redshift_table_schema(
            spark=spark,
            endpoint=ds_args.redshift_ds_config.endpoint,
            table=ds_args.redshift_ds_config.table,
            query=ds_args.redshift_ds_config.query,
            temp_s3=temp_s3,
            post_processor=batch_post_processor,
        )
    elif ds_args.HasField("snowflake_ds_config"):
        if not ds_args.snowflake_ds_config.HasField("url"):
            msg = "snowflake"
            raise core_errors.DS_ARGS_MISSING_FIELD(msg, "url")
        if not ds_args.snowflake_ds_config.HasField("database"):
            msg = "snowflake"
            raise core_errors.DS_ARGS_MISSING_FIELD(msg, "database")
        if not ds_args.snowflake_ds_config.HasField("schema"):
            msg = "snowflake"
            raise core_errors.DS_ARGS_MISSING_FIELD(msg, "schema")
        if not ds_args.snowflake_ds_config.HasField("warehouse"):
            msg = "snowflake"
            raise core_errors.DS_ARGS_MISSING_FIELD(msg, "warehouse")

        return get_snowflake_schema(
            spark=spark,
            url=ds_args.snowflake_ds_config.url,
            database=ds_args.snowflake_ds_config.database,
            schema=ds_args.snowflake_ds_config.schema,
            warehouse=ds_args.snowflake_ds_config.warehouse,
            role=ds_args.snowflake_ds_config.role if ds_args.snowflake_ds_config.HasField("role") else None,
            table=ds_args.snowflake_ds_config.table if ds_args.snowflake_ds_config.HasField("table") else None,
            query=ds_args.snowflake_ds_config.query if ds_args.snowflake_ds_config.HasField("query") else None,
            post_processor=batch_post_processor,
        )
    elif ds_args.HasField("file_ds_config"):
        schema_override = None
        if ds_args.file_ds_config.HasField("schema_override"):
            schema_override = spark_schema_wrapper.SparkSchemaWrapper.from_proto(ds_args.file_ds_config.schema_override)

        schema_uri = ds_args.file_ds_config.schema_uri if ds_args.file_ds_config.HasField("schema_uri") else None
        timestamp_column = (
            ds_args.file_ds_config.common_args.timestamp_field
            if ds_args.file_ds_config.common_args.HasField("timestamp_field")
            else None
        )
        timestamp_format = (
            ds_args.file_ds_config.timestamp_format if ds_args.file_ds_config.HasField("timestamp_format") else None
        )

        return get_file_source_schema(
            spark=spark,
            file_format=ds_args.file_ds_config.file_format,
            file_uri=ds_args.file_ds_config.uri,
            convert_to_glue=ds_args.file_ds_config.convert_to_glue_format,
            schema_uri=schema_uri,
            schema_override=schema_override,
            post_processor=batch_post_processor,
            timestamp_col=timestamp_column,
            timestmap_format=timestamp_format,
        )
    else:
        msg = f"Invalid batch source args: {ds_args}"
        raise ValueError(msg)


def derive_stream_schema(
    spark: PySparkSession,
    ds_args: virtual_data_source__args_pb2.VirtualDataSourceArgs,
    stream_post_processor: Optional[Callable],
    stream_data_source_function: Optional[Callable],
) -> spark_schema_pb2.SparkSchema:
    if ds_args.HasField("kinesis_ds_config"):
        return get_kinesis_schema(spark, ds_args.kinesis_ds_config.stream_name, stream_post_processor)
    elif ds_args.HasField("kafka_ds_config"):
        return get_kafka_schema(spark, stream_post_processor)
    elif ds_args.HasField("spark_stream_config"):
        return get_stream_data_source_function_schema(spark, stream_data_source_function)
    else:
        msg = f"Invalid stream source args: {ds_args}"
        raise ValueError(msg)


# Constructs empty data frames matching schema of DS inputs for the purpose of
# schema-validating the transformation pipeline.
def populate_empty_passed_in_inputs(
    node: PipelineNode,
    ds_map: Dict[str, specs.DataSourceSpec],
    spark: PySparkSession,
) -> Dict[str, PySparkDataFrame]:
    empty_passed_in_inputs = {}
    _populate_empty_passed_in_inputs_helper(node, empty_passed_in_inputs, ds_map, spark)
    return empty_passed_in_inputs


def _populate_empty_passed_in_inputs_helper(
    node: PipelineNode,
    empty_passed_in_inputs: Dict[str, PySparkDataFrame],
    ds_map: Dict[str, specs.DataSourceSpec],
    spark: PySparkSession,
) -> None:
    if node.HasField("data_source_node"):
        ds_id = IdHelper.to_string(node.data_source_node.virtual_data_source_id)
        ds_spec = ds_map[ds_id]
        assert (
            ds_spec.type != DataSourceType.PUSH_NO_BATCH
        ), "This utility does not support FeatureView with PushSources that do not have a batch_config"
        ds_schema = ds_spec.batch_source.spark_schema
        empty_passed_in_inputs[node.data_source_node.input_name] = spark.createDataFrame(
            [], SparkSchemaWrapper.from_proto(ds_schema).unwrap()
        )
    elif node.HasField("transformation_node"):
        for child in node.transformation_node.inputs:
            _populate_empty_passed_in_inputs_helper(child.node, empty_passed_in_inputs, ds_map, spark)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/auth/resource_role_assignments__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.auth import resource__client_pb2 as tecton__proto_dot_auth_dot_resource____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n9tecton_proto/auth/resource_role_assignments__client.proto\x12\x11tecton_proto.auth\x1a(tecton_proto/auth/resource__client.proto\"\xe8\x01\n\x1aResourceAndRoleAssignments\x12\x44\n\rresource_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x01 \x01(\tR\nresourceId\x12\x14\n\x05roles\x18\x02 \x03(\tR\x05roles\x12M\n\rroles_granted\x18\x04 \x03(\x0b\x32(.tecton_proto.auth.RoleAssignmentSummaryR\x0crolesGranted\"\xd4\x01\n\x1cResourceAndRoleAssignmentsV2\x12\x44\n\rresource_type\x18\x01 \x01(\x0e\x32\x1f.tecton_proto.auth.ResourceTypeR\x0cresourceType\x12\x1f\n\x0bresource_id\x18\x02 \x01(\tR\nresourceId\x12M\n\rroles_granted\x18\x03 \x03(\x0b\x32(.tecton_proto.auth.RoleAssignmentSummaryR\x0crolesGranted\"\xc6\x01\n\x14RoleAssignmentSource\x12N\n\x0f\x61ssignment_type\x18\x01 \x01(\x0e\x32%.tecton_proto.auth.RoleAssignmentTypeR\x0e\x61ssignmentType\x12\x30\n\x14principal_group_name\x18\x02 \x01(\tR\x12principalGroupName\x12,\n\x12principal_group_id\x18\x03 \x01(\tR\x10principalGroupId\"\x8c\x01\n\x15RoleAssignmentSummary\x12\x12\n\x04role\x18\x01 \x01(\tR\x04role\x12_\n\x17role_assignment_sources\x18\x02 \x03(\x0b\x32\'.tecton_proto.auth.RoleAssignmentSourceR\x15roleAssignmentSources*\x8a\x01\n\x12RoleAssignmentType\x12$\n ROLE_ASSIGNMENT_TYPE_UNSPECIFIED\x10\x00\x12\x1f\n\x1bROLE_ASSIGNMENT_TYPE_DIRECT\x10\x01\x12-\n)ROLE_ASSIGNMENT_TYPE_FROM_PRINCIPAL_GROUP\x10\x02\x42J\n\x0f\x63om.tecton.authP\x01Z5github.com/tecton-ai/tecton_proto/auth/resource_roles')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.auth.resource_role_assignments__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.authP\001Z5github.com/tecton-ai/tecton_proto/auth/resource_roles'
  _ROLEASSIGNMENTTYPE._serialized_start=917
  _ROLEASSIGNMENTTYPE._serialized_end=1055
  _RESOURCEANDROLEASSIGNMENTS._serialized_start=123
  _RESOURCEANDROLEASSIGNMENTS._serialized_end=355
  _RESOURCEANDROLEASSIGNMENTSV2._serialized_start=358
  _RESOURCEANDROLEASSIGNMENTSV2._serialized_end=570
  _ROLEASSIGNMENTSOURCE._serialized_start=573
  _ROLEASSIGNMENTSOURCE._serialized_end=771
  _ROLEASSIGNMENTSUMMARY._serialized_start=774
  _ROLEASSIGNMENTSUMMARY._serialized_end=914
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
A collections of builtin avro functions
"""


from pyspark import SparkContext
from pyspark.sql.column import Column, _to_java_column
from pyspark.util import _print_missing_jar


def from_avro(data, jsonFormatSchema, options={}):
    """
    Converts a binary column of Avro format into its corresponding catalyst value.
    The specified schema must match the read data, otherwise the behavior is undefined:
    it may fail or return arbitrary result.
    To deserialize the data with a compatible and evolved schema, the expected Avro schema can be
    set via the option avroSchema.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    data : :class:`~pyspark.sql.Column` or str
        the binary column.
    jsonFormatSchema : str
        the avro schema in JSON string format.
    options : dict, optional
        options to control how the Avro record is parsed.

    Notes
    -----
    Avro is built-in but external data source module since Spark 2.4. Please deploy the
    application as per the deployment section of "Apache Avro Data Source Guide".

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> from pyspark.sql.avro.functions import from_avro, to_avro
    >>> data = [(1, Row(age=2, name='Alice'))]
    >>> df = spark.createDataFrame(data, ("key", "value"))
    >>> avroDf = df.select(to_avro(df.value).alias("avro"))
    >>> avroDf.collect()
    [Row(avro=bytearray(b'\\x00\\x00\\x04\\x00\\nAlice'))]

    >>> jsonFormatSchema = '''{"type":"record","name":"topLevelRecord","fields":
    ...     [{"name":"avro","type":[{"type":"record","name":"value","namespace":"topLevelRecord",
    ...     "fields":[{"name":"age","type":["long","null"]},
    ...     {"name":"name","type":["string","null"]}]},"null"]}]}'''
    >>> avroDf.select(from_avro(avroDf.avro, jsonFormatSchema).alias("value")).collect()
    [Row(value=Row(avro=Row(age=2, name='Alice')))]
    """

    sc = SparkContext._active_spark_context
    try:
        jc = sc._jvm.org.apache.spark.sql.avro.functions.from_avro(
            _to_java_column(data), jsonFormatSchema, options)
    except TypeError as e:
        if str(e) == "'JavaPackage' object is not callable":
            _print_missing_jar("Avro", "avro", "avro", sc.version)
        raise
    return Column(jc)


def to_avro(data, jsonFormatSchema=""):
    """
    Converts a column into binary of avro format.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    data : :class:`~pyspark.sql.Column` or str
        the data column.
    jsonFormatSchema : str, optional
        user-specified output avro schema in JSON string format.

    Notes
    -----
    Avro is built-in but external data source module since Spark 2.4. Please deploy the
    application as per the deployment section of "Apache Avro Data Source Guide".

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> from pyspark.sql.avro.functions import to_avro
    >>> data = ['SPADES']
    >>> df = spark.createDataFrame(data, "string")
    >>> df.select(to_avro(df.value).alias("suite")).collect()
    [Row(suite=bytearray(b'\\x00\\x0cSPADES'))]

    >>> jsonFormatSchema = '''["null", {"type": "enum", "name": "value",
    ...     "symbols": ["SPADES", "HEARTS", "DIAMONDS", "CLUBS"]}]'''
    >>> df.select(to_avro(df.value, jsonFormatSchema).alias("suite")).collect()
    [Row(suite=bytearray(b'\\x02\\x00'))]
    """

    sc = SparkContext._active_spark_context
    try:
        if jsonFormatSchema == "":
            jc = sc._jvm.org.apache.spark.sql.avro.functions.to_avro(_to_java_column(data))
        else:
            jc = sc._jvm.org.apache.spark.sql.avro.functions.to_avro(
                _to_java_column(data), jsonFormatSchema)
    except TypeError as e:
        if str(e) == "'JavaPackage' object is not callable":
            _print_missing_jar("Avro", "avro", "avro", sc.version)
        raise
    return Column(jc)


def _test():
    import os
    import sys
    from pyspark.testing.utils import search_jar
    avro_jar = search_jar("external/avro", "spark-avro", "spark-avro")
    if avro_jar is None:
        print(
            "Skipping all Avro Python tests as the optional Avro project was "
            "not compiled into a JAR. To run these tests, "
            "you need to build Spark with 'build/sbt -Pavro package' or "
            "'build/mvn -Pavro package' before running this test.")
        sys.exit(0)
    else:
        existing_args = os.environ.get("PYSPARK_SUBMIT_ARGS", "pyspark-shell")
        jars_args = "--jars %s" % avro_jar
        os.environ["PYSPARK_SUBMIT_ARGS"] = " ".join([jars_args, existing_args])

    import doctest
    from pyspark.sql import SparkSession
    import pyspark.sql.avro.functions
    globs = pyspark.sql.avro.functions.__dict__.copy()
    spark = SparkSession.builder\
        .master("local[4]")\
        .appName("sql.avro.functions tests")\
        .getOrCreate()
    globs['spark'] = spark
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.avro.functions, globs=globs,
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



import logging
import sys

import click
from click import shell_completion

from tecton._internals import metadata_service
from tecton.cli import printer
from tecton_core import conf
from tecton_proto.data import workspace__client_pb2 as workspace_pb2
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


logger = logging.getLogger(__name__)


def create_workspace(workspace_name: str, materializable: bool):
    request = metadata_service_pb2.CreateWorkspaceRequest(
        workspace_name=workspace_name, capabilities=workspace_pb2.WorkspaceCapabilities(materializable=materializable)
    )
    metadata_service.instance().CreateWorkspace(request)
    printer.safe_print(f'Created workspace "{workspace_name}".')


def delete_workspace(workspace_name: str):
    request = metadata_service_pb2.DeleteWorkspaceRequest(workspace=workspace_name)
    metadata_service.instance().DeleteWorkspace(request)
    printer.safe_print(f'Deleted workspace "{workspace_name}".')


def get_workspace(workspace_name: str):
    request = metadata_service_pb2.GetWorkspaceRequest(workspace_name=workspace_name)
    response = metadata_service.instance().GetWorkspace(request)
    return response.workspace


def list_workspaces():
    request = metadata_service_pb2.ListWorkspacesRequest()
    response = metadata_service.instance().ListWorkspaces(request)
    return response.workspaces


def is_live_workspace(workspace_name: str) -> bool:
    request = metadata_service_pb2.GetWorkspaceRequest(workspace_name=workspace_name)
    response = metadata_service.instance().GetWorkspace(request)
    return response.workspace.capabilities.materializable


def check_workspace_exists(workspace_name: str):
    workspace_names = {w.name for w in list_workspaces()}
    if workspace_name not in workspace_names:
        printer.safe_print(
            f'Workspace "{workspace_name}" not found. Run `tecton workspace list` to see list of available workspaces.'
        )
        sys.exit(1)


class WorkspaceType(click.ParamType):
    name = "workspace"

    def shell_complete(self, ctx, param, incomplete):
        try:
            workspace_names = {w.name for w in list_workspaces()}
        except (Exception, SystemExit) as e:
            logger.error(f"\nTab-completion failed with error: {e}")
            return []

        return [shell_completion.CompletionItem(name) for name in workspace_names if name.startswith(incomplete)]


def switch_to_workspace(workspace_name: str):
    """
    Switch the selected workspace

    :param workspace_name: name of the workspace to switch to
    """
    conf.set("TECTON_WORKSPACE", workspace_name)
    conf.save_tecton_configs()
    printer.safe_print(f'Switched to workspace "{workspace_name}".')



import json
import typing
from typing import Dict
from typing import List
from typing import Optional

import pendulum

from tecton_core import errors
from tecton_core import feature_definition_wrapper
from tecton_core import id_helper
from tecton_core import materialization_context
from tecton_core import schema as core_schema
from tecton_core import specs
from tecton_core.query import pipeline_sql_builder
from tecton_core.query import sql_compat
from tecton_core.query.dialect import Dialect
from tecton_proto.args import data_source__client_pb2 as data_source_pb2
from tecton_proto.args import feature_view__client_pb2 as feature_view_pb2
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2
from tecton_proto.args import virtual_data_source__client_pb2 as virtual_data_source_pb2
from tecton_proto.common import schema__client_pb2 as schema_pb2
from tecton_proto.common import spark_schema__client_pb2 as spark_schema_pb2
from tecton_snowflake import pipeline_helper
from tecton_snowflake import snowflake_type_utils
from tecton_snowflake.utils import format_sql


if typing.TYPE_CHECKING:
    import snowflake.connector
    import snowflake.snowpark


def _get_mock_sql_for_data_source(data_source_spec: specs.DataSourceSpec) -> str:
    """Adds LIMIT 0 to data sources during fv schema derivation to improve performance.

    Similar to get_data_source_schema_sql expect this method runs on specs."""
    snowflake_source = data_source_spec.batch_source
    if type(snowflake_source) != specs.SnowflakeSourceSpec:
        msg = f"The batch source for Data Source {data_source_spec.name} must be a Snowflake source."
        raise ValueError(msg)

    if snowflake_source.table:
        sql_str = (
            f"SELECT * FROM ({snowflake_source.database}.{snowflake_source.schema}.{snowflake_source.table}) LIMIT 0"
        )
    elif snowflake_source.query:
        sql_str = f"SELECT * FROM ({snowflake_source.query}) LIMIT 0"
    else:
        msg = "A Snowflake source must have one of 'query' or 'table' set"
        raise ValueError(msg)
    return format_sql(sql_str)


def get_data_source_schema_sql(ds_args: data_source_pb2.SnowflakeDataSourceArgs) -> str:
    """Return the SQL used to query the data source provided during schema derivation.

    Similar to _get_mock_sql_for_data_source expect this method runs on data source args."""
    if ds_args.HasField("table"):
        if not ds_args.database and not ds_args.schema:
            msg = "A Snowflake source must set 'database', 'schema', and 'table' to read from a Snowflake table."
            raise ValueError(msg)
        full_table_name = f"{ds_args.database}.{ds_args.schema}.{ds_args.table}"
        sql_str = f"SELECT * FROM {full_table_name} LIMIT 0"
    elif ds_args.HasField("query"):
        sql_str = f"SELECT * FROM ({ds_args.query}) LIMIT 0"
    else:
        msg = "A Snowflake source must have one of 'query' or 'table' set"
        raise ValueError(msg)
    return format_sql(sql_str)


def get_snowflake_schema(
    ds_args: virtual_data_source_pb2.VirtualDataSourceArgs, connection: "snowflake.connector.Connection"
) -> spark_schema_pb2.SparkSchema:
    """Derive schema for snowflake data source on snowflake compute.

    This method is used for notebook driven development.
    The logic should mirror logic in resolveBatch() in SnowflakeDDL.kt.
    """
    cur = connection.cursor()

    sql_str = get_data_source_schema_sql(ds_args.snowflake_ds_config)
    try:
        cur.execute(sql_str)
    except Exception:
        msg = f"Running the following SQL failed: {sql_str}"
        raise errors.TectonInternalError(msg)

    # Get the schema from the previously ran query
    query_id = cur.sfqid
    cur.execute(f"DESCRIBE RESULT '{query_id}';")
    schema_list = cur.fetchall()  # TODO: use fetch_pandas_all() once it supports describe statements

    proto = spark_schema_pb2.SparkSchema()
    for row in schema_list:
        # schema returned is in the form (name, type,...)
        name = row[0]
        proto_field = proto.fields.add()
        proto_field.name = name
        proto_field.structfield_json = json.dumps({"name": name, "type": row[1], "nullable": True, "metadata": {}})
    return proto


def _get_mock_sql_inputs_for_schema_derivation(
    pipeline: pipeline_pb2.Pipeline, data_source_specs: List[specs.DataSourceSpec]
) -> Dict[str, str]:
    data_source_nodes = feature_definition_wrapper.pipeline_to_ds_inputs(pipeline).values()
    id_to_spec = {spec.id: spec for spec in data_source_specs}
    mock_sql_inputs = {}
    for node in data_source_nodes:
        data_source_id = id_helper.IdHelper.to_string(node.virtual_data_source_id)
        spec = id_to_spec[data_source_id]
        # Map the data source input name to the SQL to run. The input name and data source name are not always the same.
        mock_sql_inputs[node.input_name] = _get_mock_sql_for_data_source(spec)
    return mock_sql_inputs


def get_feature_view_schema_sql(
    pipeline: pipeline_pb2.Pipeline,
    transformation_specs: List[specs.TransformationSpec],
    data_source_specs: List[specs.DataSourceSpec],
    materialization_context: materialization_context.BoundMaterializationContext,
    session: Optional["snowflake.snowpark.Session"] = None,
) -> str:
    """Return the SQL used to run fv schema derivation."""
    # Create mock sql inputs for data sources to improve performance.
    mock_sql_inputs_map = _get_mock_sql_inputs_for_schema_derivation(pipeline, data_source_specs)
    if pipeline_helper.has_snowpark_transformation(transformation_specs):
        # TODO(deprecate_after=0.7): Snowpark transformations are not supported  in >= 0.8 so we can delete this and only use the PipelineSqlBuilder in the future
        return pipeline_helper.pipeline_to_sql_string(
            pipeline,
            data_source_specs,
            transformation_specs,
            mock_sql_inputs=mock_sql_inputs_map,
            materialization_context=materialization_context,
            session=session,
        )

    # maps input names to unique strings
    unique_inputs_map = {ds_input: f"{ds_input}_{id(pipeline)}" for ds_input in mock_sql_inputs_map}
    pipeline_builder = pipeline_sql_builder.PipelineSqlBuilder(
        pipeline=pipeline,
        id_to_transformation={t.id: t for t in transformation_specs},
        materialization_context=materialization_context,
        renamed_inputs_map=unique_inputs_map,
    )
    pipeline_query = pipeline_builder.get_pipeline_query(
        dialect=Dialect.SNOWFLAKE,
        input_name_to_query_map={k: sql_compat.CustomQuery(v) for k, v in mock_sql_inputs_map.items()},
    )
    return pipeline_query.get_sql()


def get_feature_view_view_schema(
    feature_view_args: feature_view_pb2.FeatureViewArgs,
    transformation_specs: List[specs.TransformationSpec],
    data_source_specs: List[specs.DataSourceSpec],
    connection: "snowflake.connector.Connection",
    session: Optional["snowflake.snowpark.Session"] = None,
) -> schema_pb2.Schema:
    """Compute the Feature View view schema.

    This method is used for notebook driven development.
    The logic should mirror logic in resolve() in SnowflakeDDL.kt.
    """
    cur = connection.cursor()

    # Create a default materialization context for the feature view.
    _tecton_materialization_context = materialization_context.BoundMaterializationContext._create_internal(
        pendulum.from_timestamp(0, pendulum.tz.UTC),
        pendulum.datetime(2100, 1, 1),
        pendulum.Duration(),
    )

    sql_str = get_feature_view_schema_sql(
        feature_view_args.pipeline, transformation_specs, data_source_specs, _tecton_materialization_context, session
    )
    try:
        cur.execute(sql_str)
    except Exception:
        msg = f"Running the following SQL failed: {sql_str}"
        raise errors.TectonInternalError(msg)

    # Get the schema from the previously ran query
    query_id = cur.sfqid
    cur.execute(f"DESCRIBE RESULT '{query_id}';")
    schema_list = cur.fetchall()  # TODO: use fetch_pandas_all() once it supports describe statements

    columns = []
    for row in schema_list:
        # schema returned is in the form (name, type,...)
        name = row[0]
        tecton_type = snowflake_type_utils.snowflake_type_to_tecton_type(row[1], name)
        column_proto = schema_pb2.Column()
        column_proto.CopyFrom(core_schema.column_from_tecton_data_type(tecton_type))
        column_proto.name = name
        columns.append(column_proto)

    return schema_pb2.Schema(columns=columns)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys

from pyspark import since, SparkContext
from pyspark.sql.column import _to_seq, _to_java_column

__all__ = ["Window", "WindowSpec"]


def _to_java_cols(cols):
    sc = SparkContext._active_spark_context
    if len(cols) == 1 and isinstance(cols[0], list):
        cols = cols[0]
    return _to_seq(sc, cols, _to_java_column)


class Window(object):
    """
    Utility functions for defining window in DataFrames.

    .. versionadded:: 1.4

    Notes
    -----
    When ordering is not defined, an unbounded window frame (rowFrame,
    unboundedPreceding, unboundedFollowing) is used by default. When ordering is defined,
    a growing window frame (rangeFrame, unboundedPreceding, currentRow) is used by default.

    Examples
    --------
    >>> # ORDER BY date ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW
    >>> window = Window.orderBy("date").rowsBetween(Window.unboundedPreceding, Window.currentRow)

    >>> # PARTITION BY country ORDER BY date RANGE BETWEEN 3 PRECEDING AND 3 FOLLOWING
    >>> window = Window.orderBy("date").partitionBy("country").rangeBetween(-3, 3)
    """

    _JAVA_MIN_LONG = -(1 << 63)  # -9223372036854775808
    _JAVA_MAX_LONG = (1 << 63) - 1  # 9223372036854775807
    _PRECEDING_THRESHOLD = max(-sys.maxsize, _JAVA_MIN_LONG)
    _FOLLOWING_THRESHOLD = min(sys.maxsize, _JAVA_MAX_LONG)

    unboundedPreceding = _JAVA_MIN_LONG

    unboundedFollowing = _JAVA_MAX_LONG

    currentRow = 0

    @staticmethod
    @since(1.4)
    def partitionBy(*cols):
        """
        Creates a :class:`WindowSpec` with the partitioning defined.
        """
        sc = SparkContext._active_spark_context
        jspec = sc._jvm.org.apache.spark.sql.expressions.Window.partitionBy(_to_java_cols(cols))
        return WindowSpec(jspec)

    @staticmethod
    @since(1.4)
    def orderBy(*cols):
        """
        Creates a :class:`WindowSpec` with the ordering defined.
        """
        sc = SparkContext._active_spark_context
        jspec = sc._jvm.org.apache.spark.sql.expressions.Window.orderBy(_to_java_cols(cols))
        return WindowSpec(jspec)

    @staticmethod
    def rowsBetween(start, end):
        """
        Creates a :class:`WindowSpec` with the frame boundaries defined,
        from `start` (inclusive) to `end` (inclusive).

        Both `start` and `end` are relative positions from the current row.
        For example, "0" means "current row", while "-1" means the row before
        the current row, and "5" means the fifth row after the current row.

        We recommend users use ``Window.unboundedPreceding``, ``Window.unboundedFollowing``,
        and ``Window.currentRow`` to specify special boundary values, rather than using integral
        values directly.

        A row based boundary is based on the position of the row within the partition.
        An offset indicates the number of rows above or below the current row, the frame for the
        current row starts or ends. For instance, given a row based sliding frame with a lower bound
        offset of -1 and a upper bound offset of +2. The frame for row with index 5 would range from
        index 4 to index 7.

        .. versionadded:: 2.1.0

        Parameters
        ----------
        start : int
            boundary start, inclusive.
            The frame is unbounded if this is ``Window.unboundedPreceding``, or
            any value less than or equal to -9223372036854775808.
        end : int
            boundary end, inclusive.
            The frame is unbounded if this is ``Window.unboundedFollowing``, or
            any value greater than or equal to 9223372036854775807.

        Examples
        --------
        >>> from pyspark.sql import Window
        >>> from pyspark.sql import functions as func
        >>> from pyspark.sql import SQLContext
        >>> sc = SparkContext.getOrCreate()
        >>> sqlContext = SQLContext(sc)
        >>> tup = [(1, "a"), (1, "a"), (2, "a"), (1, "b"), (2, "b"), (3, "b")]
        >>> df = sqlContext.createDataFrame(tup, ["id", "category"])
        >>> window = Window.partitionBy("category").orderBy("id").rowsBetween(Window.currentRow, 1)
        >>> df.withColumn("sum", func.sum("id").over(window)).sort("id", "category", "sum").show()
        +---+--------+---+
        | id|category|sum|
        +---+--------+---+
        |  1|       a|  2|
        |  1|       a|  3|
        |  1|       b|  3|
        |  2|       a|  2|
        |  2|       b|  5|
        |  3|       b|  3|
        +---+--------+---+

        """
        if start <= Window._PRECEDING_THRESHOLD:
            start = Window.unboundedPreceding
        if end >= Window._FOLLOWING_THRESHOLD:
            end = Window.unboundedFollowing
        sc = SparkContext._active_spark_context
        jspec = sc._jvm.org.apache.spark.sql.expressions.Window.rowsBetween(start, end)
        return WindowSpec(jspec)

    @staticmethod
    def rangeBetween(start, end):
        """
        Creates a :class:`WindowSpec` with the frame boundaries defined,
        from `start` (inclusive) to `end` (inclusive).

        Both `start` and `end` are relative from the current row. For example,
        "0" means "current row", while "-1" means one off before the current row,
        and "5" means the five off after the current row.

        We recommend users use ``Window.unboundedPreceding``, ``Window.unboundedFollowing``,
        and ``Window.currentRow`` to specify special boundary values, rather than using integral
        values directly.

        A range-based boundary is based on the actual value of the ORDER BY
        expression(s). An offset is used to alter the value of the ORDER BY expression, for
        instance if the current ORDER BY expression has a value of 10 and the lower bound offset
        is -3, the resulting lower bound for the current row will be 10 - 3 = 7. This however puts a
        number of constraints on the ORDER BY expressions: there can be only one expression and this
        expression must have a numerical data type. An exception can be made when the offset is
        unbounded, because no value modification is needed, in this case multiple and non-numeric
        ORDER BY expression are allowed.

        .. versionadded:: 2.1.0

        Parameters
        ----------
        start : int
            boundary start, inclusive.
            The frame is unbounded if this is ``Window.unboundedPreceding``, or
            any value less than or equal to max(-sys.maxsize, -9223372036854775808).
        end : int
            boundary end, inclusive.
            The frame is unbounded if this is ``Window.unboundedFollowing``, or
            any value greater than or equal to min(sys.maxsize, 9223372036854775807).

        Examples
        --------
        >>> from pyspark.sql import Window
        >>> from pyspark.sql import functions as func
        >>> from pyspark.sql import SQLContext
        >>> sc = SparkContext.getOrCreate()
        >>> sqlContext = SQLContext(sc)
        >>> tup = [(1, "a"), (1, "a"), (2, "a"), (1, "b"), (2, "b"), (3, "b")]
        >>> df = sqlContext.createDataFrame(tup, ["id", "category"])
        >>> window = Window.partitionBy("category").orderBy("id").rangeBetween(Window.currentRow, 1)
        >>> df.withColumn("sum", func.sum("id").over(window)).sort("id", "category").show()
        +---+--------+---+
        | id|category|sum|
        +---+--------+---+
        |  1|       a|  4|
        |  1|       a|  4|
        |  1|       b|  3|
        |  2|       a|  2|
        |  2|       b|  5|
        |  3|       b|  3|
        +---+--------+---+

        """
        if start <= Window._PRECEDING_THRESHOLD:
            start = Window.unboundedPreceding
        if end >= Window._FOLLOWING_THRESHOLD:
            end = Window.unboundedFollowing
        sc = SparkContext._active_spark_context
        jspec = sc._jvm.org.apache.spark.sql.expressions.Window.rangeBetween(start, end)
        return WindowSpec(jspec)


class WindowSpec(object):
    """
    A window specification that defines the partitioning, ordering,
    and frame boundaries.

    Use the static methods in :class:`Window` to create a :class:`WindowSpec`.

    .. versionadded:: 1.4.0
    """

    def __init__(self, jspec):
        self._jspec = jspec

    def partitionBy(self, *cols):
        """
        Defines the partitioning columns in a :class:`WindowSpec`.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        cols : str, :class:`Column` or list
            names of columns or expressions
        """
        return WindowSpec(self._jspec.partitionBy(_to_java_cols(cols)))

    def orderBy(self, *cols):
        """
        Defines the ordering columns in a :class:`WindowSpec`.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        cols : str, :class:`Column` or list
            names of columns or expressions
        """
        return WindowSpec(self._jspec.orderBy(_to_java_cols(cols)))

    def rowsBetween(self, start, end):
        """
        Defines the frame boundaries, from `start` (inclusive) to `end` (inclusive).

        Both `start` and `end` are relative positions from the current row.
        For example, "0" means "current row", while "-1" means the row before
        the current row, and "5" means the fifth row after the current row.

        We recommend users use ``Window.unboundedPreceding``, ``Window.unboundedFollowing``,
        and ``Window.currentRow`` to specify special boundary values, rather than using integral
        values directly.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        start : int
            boundary start, inclusive.
            The frame is unbounded if this is ``Window.unboundedPreceding``, or
            any value less than or equal to max(-sys.maxsize, -9223372036854775808).
        end : int
            boundary end, inclusive.
            The frame is unbounded if this is ``Window.unboundedFollowing``, or
            any value greater than or equal to min(sys.maxsize, 9223372036854775807).
        """
        if start <= Window._PRECEDING_THRESHOLD:
            start = Window.unboundedPreceding
        if end >= Window._FOLLOWING_THRESHOLD:
            end = Window.unboundedFollowing
        return WindowSpec(self._jspec.rowsBetween(start, end))

    def rangeBetween(self, start, end):
        """
        Defines the frame boundaries, from `start` (inclusive) to `end` (inclusive).

        Both `start` and `end` are relative from the current row. For example,
        "0" means "current row", while "-1" means one off before the current row,
        and "5" means the five off after the current row.

        We recommend users use ``Window.unboundedPreceding``, ``Window.unboundedFollowing``,
        and ``Window.currentRow`` to specify special boundary values, rather than using integral
        values directly.

        .. versionadded:: 1.4.0

        Parameters
        ----------
        start : int
            boundary start, inclusive.
            The frame is unbounded if this is ``Window.unboundedPreceding``, or
            any value less than or equal to max(-sys.maxsize, -9223372036854775808).
        end : int
            boundary end, inclusive.
            The frame is unbounded if this is ``Window.unboundedFollowing``, or
            any value greater than or equal to min(sys.maxsize, 9223372036854775807).
        """
        if start <= Window._PRECEDING_THRESHOLD:
            start = Window.unboundedPreceding
        if end >= Window._FOLLOWING_THRESHOLD:
            end = Window.unboundedFollowing
        return WindowSpec(self._jspec.rangeBetween(start, end))


def _test():
    import doctest
    import pyspark.sql.window
    SparkContext('local[4]', 'PythonTest')
    globs = pyspark.sql.window.__dict__.copy()
    (failure_count, test_count) = doctest.testmod(
        pyspark.sql.window, globs=globs,
        optionflags=doctest.NORMALIZE_WHITESPACE)
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



import os
import tempfile
import shutil
from pathlib import Path
import subprocess
import zipfile

PYSPARK_VERSION = "3.1.2"

parent_dir = Path(__file__).parent
vendor_dir = parent_dir / "pyspark"
shutil.rmtree(vendor_dir, ignore_errors=True)
vendor_dir.mkdir()

with tempfile.TemporaryDirectory() as d:
    d = Path(d)
    os.chdir(d)
    subprocess.check_call(["pip", "wheel", f"pyspark=={PYSPARK_VERSION}"])


    def extract_whl(package_name):
        whl_glob = list(d.glob(f"{package_name}-*.whl"))
        assert len(whl_glob) == 1
        whl = whl_glob[0]
        with zipfile.ZipFile(whl) as z:
            for name in z.namelist():
                if all((name.startswith(f"{package_name}/"),
                        name.endswith(".py"),
                        not "/tests/" in name,
                        not "/examples/" in name)):
                    print(name)
                    z.extract(name, path=vendor_dir)
        with open(vendor_dir / package_name / "BUILD.bazel", "w") as f:
            f.write("\n".join([
                "py_library(",
                "    name = \"pkg\",",
                "    srcs = glob([\"**/*.py\"]),",
                "    visibility = [\"//sdk/pypi:__pkg__\"],",
                ")",
                "",
            ]))


    extract_whl("py4j")
    extract_whl("pyspark")

subprocess.check_call(["patch", str(vendor_dir / "pyspark/context.py"), str(parent_dir / "pyspark_context.patch")])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/offlinestore/delta/transaction_writer__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import aws_credentials__client_pb2 as tecton__proto_dot_common_dot_aws__credentials____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.offlinestore.delta import metadata__client_pb2 as tecton__proto_dot_offlinestore_dot_delta_dot_metadata____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n@tecton_proto/offlinestore/delta/transaction_writer__client.proto\x12\x1ftecton_proto.offlinestore.delta\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x31tecton_proto/common/aws_credentials__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a\x36tecton_proto/offlinestore/delta/metadata__client.proto\"\xd4\x03\n\x0eInitializeArgs\x12\x12\n\x04path\x18\x01 \x01(\tR\x04path\x12\x0e\n\x02id\x18\x02 \x01(\tR\x02id\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x04 \x01(\tR\x0b\x64\x65scription\x12\x33\n\x06schema\x18\x05 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x06schema\x12+\n\x11partition_columns\x18\x06 \x03(\tR\x10partitionColumns\x12\x35\n\x17\x64ynamodb_log_table_name\x18\x07 \x01(\tR\x14\x64ynamodbLogTableName\x12\x39\n\x19\x64ynamodb_log_table_region\x18\x08 \x01(\tR\x16\x64ynamodbLogTableRegion\x12t\n\x1a\x63ross_account_role_configs\x18\t \x01(\x0b\x32\x37.tecton_proto.offlinestore.delta.CrossAccountRoleConfigR\x17\x63rossAccountRoleConfigs\x12\x1e\n\x0bkms_key_arn\x18\n \x01(\tR\tkmsKeyArn\"\xe0\x02\n\x07\x41\x64\x64\x46ile\x12\x10\n\x03uri\x18\x01 \x01(\tR\x03uri\x12h\n\x10partition_values\x18\x02 \x03(\x0b\x32=.tecton_proto.offlinestore.delta.AddFile.PartitionValuesEntryR\x0fpartitionValues\x12\x46\n\x04tags\x18\x03 \x03(\x0b\x32\x32.tecton_proto.offlinestore.delta.AddFile.TagsEntryR\x04tags\x12\x14\n\x05stats\x18\x04 \x01(\tR\x05stats\x1a\x42\n\x14PartitionValuesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xdb\x01\n\nUpdateArgs\x12\x45\n\tadd_files\x18\x02 \x03(\x0b\x32(.tecton_proto.offlinestore.delta.AddFileR\x08\x61\x64\x64\x46iles\x12Y\n\ruser_metadata\x18\x04 \x01(\x0b\x32\x34.tecton_proto.offlinestore.delta.TectonDeltaMetadataR\x0cuserMetadata\x12\x1f\n\x0b\x64\x65lete_uris\x18\x05 \x03(\tR\ndeleteUrisJ\x04\x08\x01\x10\x02J\x04\x08\x03\x10\x04\"\xad\x03\n\x0cUpdateResult\x12+\n\x11\x63ommitted_version\x18\x01 \x01(\x04R\x10\x63ommittedVersion\x12V\n\nerror_type\x18\x02 \x01(\x0e\x32\x37.tecton_proto.offlinestore.delta.UpdateResult.ErrorTypeR\terrorType\x12#\n\rerror_message\x18\x03 \x01(\tR\x0c\x65rrorMessage\"\xf2\x01\n\tErrorType\x12\x15\n\x11\x45RROR_UNSPECIFIED\x10\x00\x12\x11\n\rERROR_UNKNOWN\x10\x01\x12\x1b\n\x17\x43ONCURRENT_APPEND_ERROR\x10\x02\x12 \n\x1c\x43ONCURRENT_DELETE_READ_ERROR\x10\x03\x12\"\n\x1e\x43ONCURRENT_DELETE_DELETE_ERROR\x10\x04\x12\x1a\n\x16METADATA_CHANGED_ERROR\x10\x05\x12 \n\x1c\x43ONCURRENT_TRANSACTION_ERROR\x10\x06\x12\x1a\n\x16PROTOCOL_CHANGED_ERROR\x10\x07\"\xb7\x05\n\nExpression\x12\x35\n\x06\x63olumn\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.ColumnH\x00R\x06\x63olumn\x12O\n\x07literal\x18\x02 \x01(\x0b\x32\x33.tecton_proto.offlinestore.delta.Expression.LiteralH\x00R\x07literal\x12L\n\x06\x62inary\x18\x03 \x01(\x0b\x32\x32.tecton_proto.offlinestore.delta.Expression.BinaryH\x00R\x06\x62inary\x1a\x97\x01\n\x07Literal\x12\x12\n\x03str\x18\x01 \x01(\tH\x00R\x03str\x12:\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00R\ttimestamp\x12\x16\n\x05int64\x18\x03 \x01(\x03H\x00R\x05int64\x12\x14\n\x04\x62ool\x18\x04 \x01(\x08H\x00R\x04\x62oolB\x0e\n\x0cliteral_type\x1a\xa5\x02\n\x06\x42inary\x12\x45\n\x02op\x18\x01 \x01(\x0e\x32\x35.tecton_proto.offlinestore.delta.Expression.Binary.OpR\x02op\x12?\n\x04left\x18\x02 \x01(\x0b\x32+.tecton_proto.offlinestore.delta.ExpressionR\x04left\x12\x41\n\x05right\x18\x03 \x01(\x0b\x32+.tecton_proto.offlinestore.delta.ExpressionR\x05right\"P\n\x02Op\x12\x12\n\x0eOP_UNSPECIFIED\x10\x00\x12\n\n\x06OP_AND\x10\x01\x12\t\n\x05OP_LT\x10\x02\x12\t\n\x05OP_LE\x10\x03\x12\t\n\x05OP_EQ\x10\x04\x12\t\n\x05OP_OR\x10\x05\x42\x11\n\x0f\x65xpression_type\"g\n\x11ReadForUpdateArgs\x12R\n\x0eread_predicate\x18\x01 \x01(\x0b\x32+.tecton_proto.offlinestore.delta.ExpressionR\rreadPredicate\")\n\x13ReadForUpdateResult\x12\x12\n\x04uris\x18\x01 \x03(\tR\x04uris\"\x9e\x01\n\x11GetPartitionsArgs\x12P\n\x04tags\x18\x01 \x03(\x0b\x32<.tecton_proto.offlinestore.delta.GetPartitionsArgs.TagsEntryR\x04tags\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xa2\x02\n\x13GetPartitionsResult\x12^\n\npartitions\x18\x01 \x03(\x0b\x32>.tecton_proto.offlinestore.delta.GetPartitionsResult.PartitionR\npartitions\x1a\xaa\x01\n\tPartition\x12\x62\n\x06values\x18\x01 \x03(\x0b\x32J.tecton_proto.offlinestore.delta.GetPartitionsResult.Partition.ValuesEntryR\x06values\x1a\x39\n\x0bValuesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xc8\x01\n\x16\x43rossAccountRoleConfig\x12R\n\x15s3_cross_account_role\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.common.AwsIamRoleR\x12s3CrossAccountRole\x12Z\n\x19\x64ynamo_cross_account_role\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.common.AwsIamRoleR\x16\x64ynamoCrossAccountRoleB9\n\x1d\x63om.tecton.offlinestore.deltaB\x16TransactionWriterProtoP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.offlinestore.delta.transaction_writer__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.tecton.offlinestore.deltaB\026TransactionWriterProtoP\001'
  _ADDFILE_PARTITIONVALUESENTRY._options = None
  _ADDFILE_PARTITIONVALUESENTRY._serialized_options = b'8\001'
  _ADDFILE_TAGSENTRY._options = None
  _ADDFILE_TAGSENTRY._serialized_options = b'8\001'
  _GETPARTITIONSARGS_TAGSENTRY._options = None
  _GETPARTITIONSARGS_TAGSENTRY._serialized_options = b'8\001'
  _GETPARTITIONSRESULT_PARTITION_VALUESENTRY._options = None
  _GETPARTITIONSRESULT_PARTITION_VALUESENTRY._serialized_options = b'8\001'
  _INITIALIZEARGS._serialized_start=284
  _INITIALIZEARGS._serialized_end=752
  _ADDFILE._serialized_start=755
  _ADDFILE._serialized_end=1107
  _ADDFILE_PARTITIONVALUESENTRY._serialized_start=984
  _ADDFILE_PARTITIONVALUESENTRY._serialized_end=1050
  _ADDFILE_TAGSENTRY._serialized_start=1052
  _ADDFILE_TAGSENTRY._serialized_end=1107
  _UPDATEARGS._serialized_start=1110
  _UPDATEARGS._serialized_end=1329
  _UPDATERESULT._serialized_start=1332
  _UPDATERESULT._serialized_end=1761
  _UPDATERESULT_ERRORTYPE._serialized_start=1519
  _UPDATERESULT_ERRORTYPE._serialized_end=1761
  _EXPRESSION._serialized_start=1764
  _EXPRESSION._serialized_end=2459
  _EXPRESSION_LITERAL._serialized_start=1993
  _EXPRESSION_LITERAL._serialized_end=2144
  _EXPRESSION_BINARY._serialized_start=2147
  _EXPRESSION_BINARY._serialized_end=2440
  _EXPRESSION_BINARY_OP._serialized_start=2360
  _EXPRESSION_BINARY_OP._serialized_end=2440
  _READFORUPDATEARGS._serialized_start=2461
  _READFORUPDATEARGS._serialized_end=2564
  _READFORUPDATERESULT._serialized_start=2566
  _READFORUPDATERESULT._serialized_end=2607
  _GETPARTITIONSARGS._serialized_start=2610
  _GETPARTITIONSARGS._serialized_end=2768
  _GETPARTITIONSARGS_TAGSENTRY._serialized_start=1052
  _GETPARTITIONSARGS_TAGSENTRY._serialized_end=1107
  _GETPARTITIONSRESULT._serialized_start=2771
  _GETPARTITIONSRESULT._serialized_end=3061
  _GETPARTITIONSRESULT_PARTITION._serialized_start=2891
  _GETPARTITIONSRESULT_PARTITION._serialized_end=3061
  _GETPARTITIONSRESULT_PARTITION_VALUESENTRY._serialized_start=3004
  _GETPARTITIONSRESULT_PARTITION_VALUESENTRY._serialized_end=3061
  _CROSSACCOUNTROLECONFIG._serialized_start=3064
  _CROSSACCOUNTROLECONFIG._serialized_end=3264
# @@protoc_insertion_point(module_scope)



from typing import Union

# The `pyspark.sql.connect.*` classes are the default for the Notebooks on Databricks Shared Access Mode Clusters of
# Version 14.3 or higher. That's why we need to support them.
from pyspark.sql import SparkSession


try:
    from pyspark.sql.connect.session import SparkSession as ConnectSparkSession
except ImportError:
    ConnectSparkSession = SparkSession

from pyspark.sql.dataframe import DataFrame


try:
    from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame
except ImportError:
    ConnectDataFrame = DataFrame

from pyspark.sql import Column


try:
    from pyspark.sql.connect.column import Column as ConnectColumn
except ImportError:
    ConnectColumn = Column

PySparkSession = Union[SparkSession, ConnectSparkSession]
PySparkDataFrame = Union[DataFrame, ConnectDataFrame]
PySparkColumn = Union[Column, ConnectColumn]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/aws_credentials__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1tecton_proto/common/aws_credentials__client.proto\x12\x13tecton_proto.common\x1a\x1fgoogle/protobuf/timestamp.proto\"\xc1\x01\n\x0e\x41wsCredentials\x12\"\n\raccess_key_id\x18\x01 \x01(\tR\x0b\x61\x63\x63\x65ssKeyId\x12*\n\x11secret_access_key\x18\x02 \x01(\tR\x0fsecretAccessKey\x12#\n\rsession_token\x18\x03 \x01(\tR\x0csessionToken\x12:\n\nexpiration\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\nexpiration\"\x96\x01\n\nAwsIamRole\x12\x19\n\x08role_arn\x18\x01 \x01(\tR\x07roleArn\x12L\n\x11intermediate_role\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.common.AwsIamRoleR\x10intermediateRole\x12\x1f\n\x0b\x65xternal_id\x18\x03 \x01(\tR\nexternalIdB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.aws_credentials__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _AWSCREDENTIALS._serialized_start=108
  _AWSCREDENTIALS._serialized_end=301
  _AWSIAMROLE._serialized_start=304
  _AWSIAMROLE._serialized_end=454
# @@protoc_insertion_point(module_scope)



from tecton_core import conf
from tecton_core.query import nodes
from tecton_core.query.duckdb import nodes as duckdb_nodes
from tecton_core.query.node_interface import NodeRef


class DuckDBTreeRewriter:
    node_mapping = {
        nodes.PartialAggNode: duckdb_nodes.PartialAggDuckDBNode,
        nodes.AsofJoinFullAggNode: duckdb_nodes.AsofJoinFullAggNodeDuckDBNode,
        nodes.AsofJoinNode: duckdb_nodes.AsofJoinDuckDBNode,
    }

    def rewrite(
        self,
        tree: NodeRef,
    ) -> None:
        for i in tree.inputs:
            self.rewrite(tree=i)
        tree_node = tree.node
        if isinstance(tree_node, nodes.AsofJoinFullAggNode) and conf.get_bool("DUCKDB_ENABLE_OPTIMIZED_FULL_AGG"):
            tree.node = duckdb_nodes.AsofJoinFullAggDuckDBNodeV2.from_query_node(tree_node)
        elif tree_node.__class__ in self.node_mapping:
            tree.node = self.node_mapping[tree_node.__class__].from_query_node(tree_node)



"""
We vendor dill because it's a very important dependency to keep pinned (since
it handle serialization/deserialization) and it can conflict with other
dependencies users have (Beam is a known conflict).

Dill is a python-only package so it is easy to vendor fully.
"""

import os
import tempfile
import shutil
from pathlib import Path
import subprocess
import zipfile

DILL_VERSION = "0.3.0"

parent_dir = Path(__file__).parent
vendor_dir = parent_dir / "dill"
shutil.rmtree(vendor_dir, ignore_errors=True)
vendor_dir.mkdir()

with tempfile.TemporaryDirectory() as d:
    d = Path(d)
    os.chdir(d)
    subprocess.check_call(["pip", "wheel", f"dill=={DILL_VERSION}"])


    def extract_whl(package_name):
        whl_glob = list(d.glob(f"{package_name}-*.whl"))
        assert len(whl_glob) == 1
        whl = whl_glob[0]
        with zipfile.ZipFile(whl) as z:
            for name in z.namelist():
                if all((name.startswith(f"{package_name}/"),
                        name.endswith(".py"),
                        not "/tests/" in name,
                        not "/examples/" in name)):
                    print(name)
                    z.extract(name, path=vendor_dir)

    extract_whl("dill")


subprocess.check_call(["patch", str(vendor_dir / "dill/objtypes.py"), str(parent_dir / "dill_objtypes.patch")])
